#include "benchmarker.h"

#include <unistd.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include "app_error.h"
#include "tcp_client.h"
#include "scan_config.h"
#include "request_response.h"
#include "state_machine.h"
#include "log.h"


static float gpu_execution[N_SAMPLES];
static float tcp_execution[N_SAMPLES];
static float tot_execution[N_SAMPLES];

static uint32_t samples_collected;
static uint32_t current_image_buffer_size;
static uint32_t current_block_size;

static float find_min(float * p_data, uint32_t len)
{
	uint32_t i;
	float min = 100000000000; // large initial number to suppress 0 values

	for (i = 0; i < len; i++)
	{
		if (p_data[i] < min && p_data[i] != 0)
		{
			min = p_data[i];
		}
	}
	return min;
}

static float find_max(float * p_data, uint32_t len)
{
	uint32_t i;
	float max = p_data[0];

	for (i = 0; i < len; i++)
	{
		if (p_data[i] > max)
		{
			max = p_data[i];
		}
	}
	return max;
}

static float find_dev(float * p_data, float av, uint32_t len)
{
	uint32_t i;
	float var = 0.0;

	for (i = 0; i < len; i++)
	{
		if (p_data[i] != 0)
		{
			var += (p_data[i] - av)*(p_data[i] - av);
		}
	}

	var = var/(len - 1);

	return var;
}

static float find_av(float * p_data, uint32_t len)
{
	float av = 0;
	uint32_t i;

	for (i = 0; i < len; i++)
	{
		if (p_data[i] != 0)
		{
			av += p_data[i]/len;
		}
	}
	return av;
}


static ret_code_t save_samples(uint32_t block_size, uint32_t image_buffer_size)
{
	float gpu_av  = find_av (gpu_execution, N_SAMPLES);
	float gpu_min = find_min(gpu_execution, N_SAMPLES);
	float gpu_max = find_max(gpu_execution, N_SAMPLES);
	float gpu_dev = find_dev(gpu_execution, gpu_av, N_SAMPLES);

	float tcp_av  = find_av (tcp_execution, N_SAMPLES);
	float tcp_min = find_min(tcp_execution, N_SAMPLES);
	float tcp_max = find_max(tcp_execution, N_SAMPLES);
	float tcp_dev = find_dev(tcp_execution, tcp_av, N_SAMPLES);

	float tot_av  = find_av (tot_execution, N_SAMPLES);
	float tot_min = find_min(tot_execution, N_SAMPLES);
	float tot_max = find_max(tot_execution, N_SAMPLES);
	float tot_dev = find_dev(tot_execution, tot_av, N_SAMPLES);

	FILE * fp;
	uint32_t i;
	static char current_time_string[128] = {};
	char filename_buf[256];

	// obtain time only first time running
	if (current_time_string[0] == '\0')
	{
		time_t t = time(NULL);
		struct tm *tm = localtime(&t);
		strftime(current_time_string, sizeof(current_time_string), "%c", tm);
		LOG("Samples are saved with timestamp: %s\n", current_time_string);
	}

	// replace : with _ in the current_time_string, because it is not compatible with windows file names
	for (i = 0; i < strlen(current_time_string); i++)
	{
		if (current_time_string[i] == ':')
		{
			current_time_string[i] = '_';
		}
	}

	sprintf(filename_buf, "../../BenchmarkSamples/%s ImgBuf%dBlockSize%d.txt",
							current_time_string,
							image_buffer_size,
							block_size);

	fp = fopen(filename_buf, "w");
	if (fp == NULL)
	{
		perror("Failed opening file");
		return RET_ERROR_NO_MEM;
	}

	fprintf(fp, "MIN, MAX, AV, DEV\n");
	fprintf(fp, "GPU: %f, %f, %f, %f\n", gpu_min, gpu_max, gpu_av, gpu_dev);
	fprintf(fp, "TCP: %f, %f, %f, %f\n", tcp_min, tcp_max, tcp_av, tcp_dev);
	fprintf(fp, "TOT: %f, %f, %f, %f\n", tot_min, tot_max, tot_av, tot_dev);

	fprintf(fp, "Raw data for gpu execution time:\n");
	for (i = 0; i < N_SAMPLES; i++)
	{
		fprintf(fp, "%f\n", gpu_execution[i]);
	}
	fprintf(fp, "Raw data for tcp execution time:\n");
	for (i = 0; i < N_SAMPLES; i++)
	{
		fprintf(fp, "%f\n", tcp_execution[i]);
	}
	fprintf(fp, "Raw data for total execution time:\n");
	for (i = 0; i < N_SAMPLES; i++)
	{
		fprintf(fp, "%f\n", tot_execution[i]);
	}

	return RET_SUCCESS;
}



static ret_code_t send_scan_config(uint32_t block_size, uint32_t image_buffer_size)
{
	scan_config_t config;
	ret_code_t    ret_code;
	config.scan_mode = SCAN_MODE_4;

	config.image_info.frame_height 			   = HEIGHT;
	config.image_info.num_scan_lines_per_image = SCAN_LINES_PER_IMAGE;
	config.scanner_ready_pin                   = gpio57; //gpio36 ;
	config.fps								   = 200;
	config.cycle_time_diff_threshold_ms        = 0.2;

	config.img_processing_config.threads_per_block  = block_size;
	config.img_processing_config.frames_per_message = image_buffer_size;
	config.img_processing_config.num_streams 	    = N_STREAMS;
	config.img_processing_config.scatter_distance   = 10;
	config.img_processing_config.threshold_value    = 0;
	config.img_processing_config.pixel_power        = 3;

	message_t request, response;
	request.type = MSG_TYPE_CONFIG_SCAN;
	request.request.p_payload = (void*)&config;
	request.request.payload_len = sizeof(config);

	ret_code = issue_request(&request, &response);
	if (ret_code != RET_SUCCESS)
	{
		LOG("Failed to issue request\n");
		return ret_code;
	}
	if (response.response.ret_code != RET_SUCCESS)
	{
		LOG("Response did not return success\n");
		return ret_code;
	}

	return RET_SUCCESS;
}



void benchmarker_do_all(void)
{
	const uint32_t block_sizes[] 	    = BLOCK_SIZES;
	const uint32_t image_buffer_sizes[] = IMAGE_BUFFER_SIZES;

	uint32_t i, j;
	ret_code_t ret_code;

	ret_code = state_machine_goto_state(STATE_IDLE);
	APP_ERROR_CHECK(ret_code);

	for (i = 0; i < sizeof(image_buffer_sizes)/sizeof(image_buffer_sizes[0]); i++)
	{
		for (j = 0; j < sizeof(block_sizes)/sizeof(block_sizes[0]); j++)
		{
			LOG("Current Block size: %d, image buffer size: %d\n", block_sizes[j], image_buffer_sizes[i]);

			benchmarker_start(block_sizes[j], image_buffer_sizes[i]);
		}
	}

}

void benchmarker_start(uint32_t block_size, uint32_t image_buffer_size)
{
	ret_code_t ret_code;

	current_image_buffer_size = image_buffer_size;
	current_block_size        = block_size;

	ret_code = state_machine_goto_state(STATE_CONFIG);
	APP_ERROR_CHECK(ret_code);

	ret_code = send_scan_config(block_size, image_buffer_size);
	APP_ERROR_CHECK(ret_code);

	samples_collected = 0;

	ret_code = state_machine_goto_state(STATE_SCANNING);
	APP_ERROR_CHECK(ret_code);

	while (samples_collected < N_SAMPLES)
	{
		sleep(1);
	}

	ret_code = state_machine_goto_state(STATE_IDLE);
	APP_ERROR_CHECK(ret_code);

	ret_code = save_samples(block_size, image_buffer_size);
	APP_ERROR_CHECK(ret_code);

}


static ret_code_t set_buffer_pointers(message_t * p_message, scan_output_buffer_t * p_scan_output_buffer)
{
	const uint32_t SCAN_LINES = SCAN_LINES_PER_IMAGE * current_image_buffer_size;

    // first check size
	p_scan_output_buffer->buffer_size = 0;
	p_scan_output_buffer->buffer_size += sizeof(scan_config_t);
	p_scan_output_buffer->buffer_size += sizeof(*p_scan_output_buffer->p_height_buffer) * SCAN_LINES;
	p_scan_output_buffer->buffer_size += sizeof(*p_scan_output_buffer->p_intensity_buffer) * SCAN_LINES;
	p_scan_output_buffer->buffer_size += sizeof(*p_scan_output_buffer->p_reflectance_buffer) * SCAN_LINES;
	p_scan_output_buffer->buffer_size += sizeof(*p_scan_output_buffer->p_scatter_buffer) * SCAN_LINES;
	p_scan_output_buffer->buffer_size += sizeof(*p_scan_output_buffer->p_rgb_buffer) * SCAN_LINES * 3;
	p_scan_output_buffer->buffer_size += sizeof(*p_scan_output_buffer->p_gpu_execution_time);
	p_scan_output_buffer->buffer_size += sizeof(*p_scan_output_buffer->p_total_execution_time);

	if (p_scan_output_buffer->buffer_size != p_message->request.payload_len)
	{
		LOG("Wrong message length received. Got: %d, expected %d\n",
				p_message->request.payload_len,
				p_scan_output_buffer->buffer_size);
		return RET_ERROR_INVALID_LENGTH;
	}

	// set pointers
	uint32_t index = 0;
	p_scan_output_buffer->p_config         = (scan_config_t *)&p_message->request.p_payload[index];
	index += sizeof(scan_config_t);

	p_scan_output_buffer->p_height_buffer         = (uint16_t*)&p_message->request.p_payload[index];
	index += sizeof(*p_scan_output_buffer->p_height_buffer) * SCAN_LINES;

	p_scan_output_buffer->p_intensity_buffer         = (uint16_t*)&p_message->request.p_payload[index];
	index += sizeof(*p_scan_output_buffer->p_intensity_buffer) * SCAN_LINES;

	p_scan_output_buffer->p_scatter_buffer         = (uint16_t*)&p_message->request.p_payload[index];
	index += sizeof(*p_scan_output_buffer->p_scatter_buffer) * SCAN_LINES;

	p_scan_output_buffer->p_reflectance_buffer         = (uint16_t*)&p_message->request.p_payload[index];
	index += sizeof(*p_scan_output_buffer->p_reflectance_buffer) * SCAN_LINES;

	p_scan_output_buffer->p_rgb_buffer         = (uint16_t*)&p_message->request.p_payload[index];
	index += sizeof(*p_scan_output_buffer->p_rgb_buffer) * SCAN_LINES * 3;

	p_scan_output_buffer->p_gpu_execution_time         = (float*)&p_message->request.p_payload[index];
	index += sizeof(*p_scan_output_buffer->p_gpu_execution_time);

	p_scan_output_buffer->p_total_execution_time         = (float*)&p_message->request.p_payload[index];

	return RET_SUCCESS;
}

void benchmarker_on_scan_data(message_t * p_message)
{
	if (samples_collected >= N_SAMPLES)
	{
		return;
	}

	ret_code_t 			 ret_code;
	scan_output_buffer_t scan_output_buffer;

	ret_code = set_buffer_pointers(p_message, &scan_output_buffer);
	APP_ERROR_CHECK(ret_code);

	gpu_execution[samples_collected] = *scan_output_buffer.p_gpu_execution_time;
	tot_execution[samples_collected] = *scan_output_buffer.p_total_execution_time;
	tcp_execution[samples_collected] = tot_execution[samples_collected] - gpu_execution[samples_collected];

	samples_collected++;

}
