#ifndef BENCHMARKER_H__
#define BENCHMARKER_H__

#include "tcp_client.h"
#include "scan_config.h"

/** @brief The benchmark configuration parameters. */
#define N_SAMPLES   		 300
#define SCAN_LINES_PER_IMAGE 960
#define HEIGHT				 192
#define N_STREAMS   		 4
#define BLOCK_SIZES			 {32,64,128,256,512}
#define IMAGE_BUFFER_SIZES	 {1,2,4,8,16,32,64,128}


/** @brief A structure holding a pointer to the output buffer.
 *
 *  @details    The structure does also hold pointers to sub-buffers of the output buffer.
 *              These pointers are only calculated once, reducing the computational load.
 * */
typedef struct scan_output_buffer_s
{
    uint8_t                 * p_buffer;                 /**< Pointer to the raw output buffer. */
    uint32_t                buffer_size;                /**< The size of the output buffer in bytes. */

    scan_config_t           * p_config;                 /**< Pointer to the configuration field of the output buffer. */

    uint16_t               * p_height_buffer;           /**< Height is stored in a short word. The field is num_scan_lines_per_image*frames_per_message long. */
    uint16_t               * p_intensity_buffer;        /**< Pointer to the location where intensity data is stored. It represents the intensity of the laser line. */
    uint16_t               * p_scatter_buffer;          /**< The scatter is the intensity some distance from the laser line. */
    uint16_t               * p_reflectance_buffer;      /**< The reflectance is the average intensity of a scan column. */
    uint16_t               * p_rgb_buffer;              /**< The RGB buffer stores the RGB values in a 16-bit triplet. */
    float                  * p_gpu_execution_time;      /**< The previous GPU execution time measured in milliseconds. */
    float                  * p_total_execution_time;    /**< The previous GPU + TCP execution time measured in milliseconds. */
} scan_output_buffer_t;


/** @brief The benchmarker must be called when scan data is received.
 *
 *  @details The benchmark parameters are extracted from the scan data.
 *
 *  @param[in] p_message	The scan data message.
 */
void benchmarker_on_scan_data(message_t * p_message);


/** @brief Starts the benchmarker for all image buffer and thread block sizes.
 */
void benchmarker_do_all(void);


/** @brief Starts the benchmarker with a given configuration.
 */
void benchmarker_start(uint32_t block_size, uint32_t image_buffer_size);


#endif // BENCHMARKER_H__
