#include "camera_config.h"

#include <string.h>
#include "request_response.h"
#include "app_error.h"
#include "log.h"


typedef struct
{
	fc2Config 				config;
	fc2Property 			properties[FC2_UNSPECIFIED_PROPERTY_TYPE];
	fc2VideoMode 			videoMode;
	fc2FrameRate 			frameRate;
	fc2Format7ImageSettings format7ImageSettings;
	unsigned int 			format7PacketSize;
	float                  	format7PacketPercentSpeed;
	fc2EmbeddedImageInfo    embeddedImageInfo;
	fc2TriggerMode		    triggerMode;
	fc2TriggerDelay			triggerDelay;
} camera_state_t;

static camera_state_t m_camera_state;


float camera_check_max_frame_rate(uint32_t width, uint32_t height, fc2PixelFormat pixel_format)
{
	ret_code_t ret_code;

	// load from code
	m_camera_state.config.asyncBusSpeed 			= FC2_BUSSPEED_ANY;
	m_camera_state.config.bandwidthAllocation 		= FC2_BANDWIDTH_ALLOCATION_UNSUPPORTED;
	m_camera_state.config.grabMode 					= FC2_BUFFER_FRAMES;
	m_camera_state.config.grabTimeout 				= -1;
#ifdef FLYCAPTURE2_SDK_1604
	m_camera_state.config.highPerformanceRetrieveBuffer = 1;
#endif
	m_camera_state.config.isochBusSpeed 			= FC2_BUSSPEED_S5000;
	m_camera_state.config.minNumImageNotifications 	= 1;
	m_camera_state.config.numBuffers 				= 32;
	m_camera_state.config.numImageNotifications 	= 1;

	m_camera_state.videoMode = FC2_VIDEOMODE_FORMAT7;
	m_camera_state.frameRate = FC2_FRAMERATE_FORMAT7;

	m_camera_state.format7ImageSettings.mode 		= FC2_MODE_0;
	m_camera_state.format7ImageSettings.offsetX 	= 0;
	m_camera_state.format7ImageSettings.offsetY 	= 0;
	m_camera_state.format7ImageSettings.width 		= width;
	m_camera_state.format7ImageSettings.height 		= height;
	m_camera_state.format7ImageSettings.pixelFormat = pixel_format;

	message_t request, response;
	camera_config_t config;

	request.type = MSG_TYPE_CONFIG_CAMERA;
	request.request.p_payload = (void*)&config;

	config.type = CAMERA_CONFIG_fc2SetConfiguration;
	memcpy(&config.data.config, &m_camera_state.config, sizeof(m_camera_state.config));
	request.request.payload_len = 4 + sizeof(m_camera_state.config);
	ret_code = issue_request(&request, &response);

	if (ret_code != RET_SUCCESS)
	{
		LOG("Failed issueing request, %d\n", ret_code);
		return 0;
	}
	if (response.response.ret_code != RET_SUCCESS)
	{
		LOG("Failed setting configuration, %d\n", response.response.ret_code);
		return 0;
	}


    config.type = CAMERA_CONFIG_fc2ValidateFormat7Settings;

    memcpy(&config.data.format7Settings.imageSettings,
		   &m_camera_state.format7ImageSettings,
		   sizeof(m_camera_state.format7ImageSettings));
    ret_code = camera_config(&config, 4 + sizeof(config.data.format7Settings.imageSettings));
    if (ret_code == RET_SUCCESS)
    {
    	m_camera_state.format7PacketSize = config.data.format7Settings.packetInfo.recommendedBytesPerPacket;
    	m_camera_state.format7PacketPercentSpeed = 100.0;
    }
    else
    {
    	LOG("FAILED validating format 7 info\n");
    	return 0;
    }


	config.type = CAMERA_CONFIG_fc2SetFormat7ConfigurationPacket;
	memcpy(&config.data.format7Configuration.imageSettings,
			&m_camera_state.format7ImageSettings,
				sizeof(m_camera_state.format7ImageSettings));
	config.data.format7Configuration.packetSize = m_camera_state.format7PacketSize;
	request.request.p_payload = (void*)&config;
	request.request.payload_len = 4 + sizeof(m_camera_state.format7ImageSettings) + sizeof(m_camera_state.format7PacketSize);
	ret_code = issue_request(&request, &response);

	if (ret_code != RET_SUCCESS)
	{
		LOG("Failed issueing request, %d\n", ret_code);
		return 0;
	}
	if (response.response.ret_code != RET_SUCCESS)
	{
		LOG("Failed setting format 7 configuration packet request, %d\n", response.response.ret_code);
		return 0;
	}


    config.type = CAMERA_CONFIG_fc2GetPropertyInfo;
    config.data.propertyInfo.type = FC2_FRAME_RATE;
    ret_code = camera_config(&config, 4 + sizeof(config.data.propertyInfo.type));
    if (ret_code != RET_SUCCESS)
    {
    	LOG("FAILED getting framerate property info\n");
    	return 0;
    }

	return config.data.propertyInfo.absMax;
}

ret_code_t camera_config_send_all(void)
{
	ret_code_t ret_code;

	// load from code
	m_camera_state.config.asyncBusSpeed 			= FC2_BUSSPEED_ANY;
	m_camera_state.config.bandwidthAllocation 		= FC2_BANDWIDTH_ALLOCATION_UNSUPPORTED;
	m_camera_state.config.grabMode 					= FC2_BUFFER_FRAMES;
	m_camera_state.config.grabTimeout 				= -1;
#ifdef FLYCAPTURE2_SDK_1604
	m_camera_state.config.highPerformanceRetrieveBuffer = 1;
#endif
	m_camera_state.config.isochBusSpeed 			= FC2_BUSSPEED_S5000;
	m_camera_state.config.minNumImageNotifications 	= 1;
	m_camera_state.config.numBuffers 				= 32;
	m_camera_state.config.numImageNotifications 	= 1;

	m_camera_state.videoMode = FC2_VIDEOMODE_FORMAT7;
	m_camera_state.frameRate = FC2_FRAMERATE_FORMAT7;

	m_camera_state.format7ImageSettings.mode 		= FC2_MODE_0;
	m_camera_state.format7ImageSettings.offsetX 	= 0;
	m_camera_state.format7ImageSettings.offsetY 	= 0;
	m_camera_state.format7ImageSettings.width 		= 1920;
	m_camera_state.format7ImageSettings.height 		= 192;
	m_camera_state.format7ImageSettings.pixelFormat = FC2_PIXEL_FORMAT_RAW16;

	m_camera_state.format7PacketPercentSpeed = 100.0;

	m_camera_state.properties[FC2_FRAME_RATE].type				= FC2_FRAME_RATE;
	m_camera_state.properties[FC2_FRAME_RATE].absControl		= 1;
	m_camera_state.properties[FC2_FRAME_RATE].onOff				= 1;
	m_camera_state.properties[FC2_FRAME_RATE].autoManualMode	= 0;
	m_camera_state.properties[FC2_FRAME_RATE].valueA			= 1;
	m_camera_state.properties[FC2_FRAME_RATE].absValue			= 10; // Set value here
	m_camera_state.properties[FC2_FRAME_RATE].onePush			= 0;

	m_camera_state.properties[FC2_GAIN].type			= FC2_GAIN;
	m_camera_state.properties[FC2_GAIN].absControl		= 1;
	m_camera_state.properties[FC2_GAIN].onOff			= 1;
	m_camera_state.properties[FC2_GAIN].autoManualMode	= 0;
	m_camera_state.properties[FC2_GAIN].valueA			= 1;
	m_camera_state.properties[FC2_GAIN].absValue		= 29.996418; // Set value here
	m_camera_state.properties[FC2_GAIN].onePush			= 0;

	m_camera_state.properties[FC2_SHUTTER].type				= FC2_SHUTTER;
	m_camera_state.properties[FC2_SHUTTER].absControl		= 1;
	m_camera_state.properties[FC2_SHUTTER].onOff			= 1;
	m_camera_state.properties[FC2_SHUTTER].autoManualMode	= 0;
	m_camera_state.properties[FC2_SHUTTER].valueA			= 1;
	m_camera_state.properties[FC2_SHUTTER].absValue			= 0.5; // Set value here
	m_camera_state.properties[FC2_SHUTTER].onePush			= 0;

	m_camera_state.properties[FC2_TRIGGER_DELAY].type		    = FC2_TRIGGER_DELAY;
	m_camera_state.properties[FC2_TRIGGER_DELAY].absControl		= 1;
	m_camera_state.properties[FC2_TRIGGER_DELAY].onOff			= 1;
	m_camera_state.properties[FC2_TRIGGER_DELAY].autoManualMode	= 0;
	m_camera_state.properties[FC2_TRIGGER_DELAY].valueA			= 1;
	m_camera_state.properties[FC2_TRIGGER_DELAY].absValue	    = 0.0005; // Set value here
	m_camera_state.properties[FC2_TRIGGER_DELAY].onePush		= 0;

	m_camera_state.properties[FC2_AUTO_EXPOSURE].type			= FC2_AUTO_EXPOSURE;
	m_camera_state.properties[FC2_AUTO_EXPOSURE].absControl		= 1;
	m_camera_state.properties[FC2_AUTO_EXPOSURE].onOff			= 0;
	m_camera_state.properties[FC2_AUTO_EXPOSURE].autoManualMode	= 0;
	m_camera_state.properties[FC2_AUTO_EXPOSURE].valueA			= 1;
	m_camera_state.properties[FC2_AUTO_EXPOSURE].absValue		= 0.3; // Set value here
	m_camera_state.properties[FC2_AUTO_EXPOSURE].onePush		= 0;

	m_camera_state.triggerMode.onOff = 1;
	m_camera_state.triggerMode.source = 0;
	m_camera_state.triggerMode.mode = 14;
	m_camera_state.triggerMode.parameter = 1;

	memset(&m_camera_state.embeddedImageInfo, 0, sizeof(m_camera_state.embeddedImageInfo));
	m_camera_state.embeddedImageInfo.GPIOPinState.onOff = 1;
	m_camera_state.embeddedImageInfo.timestamp.onOff    = 1;

	message_t request, response;
	camera_config_t cam_config;

	request.type = MSG_TYPE_CONFIG_CAMERA;
	request.request.p_payload = (void*)&cam_config;

	LOG_DEBUG("Calling fc2SetConfiguration\n");
	cam_config.type = CAMERA_CONFIG_fc2SetConfiguration;
	memcpy(&cam_config.data.config, &m_camera_state.config, sizeof(m_camera_state.config));
	request.request.payload_len = 4 + sizeof(m_camera_state.config);
	ret_code = issue_request(&request, &response);
	RETURN_ON_ERROR(ret_code);
	RETURN_ON_ERROR(response.response.ret_code);

    cam_config.type = CAMERA_CONFIG_fc2ValidateFormat7Settings;

    memcpy(&cam_config.data.format7Settings.imageSettings, &m_camera_state.format7ImageSettings, sizeof(m_camera_state.format7ImageSettings));
    ret_code = camera_config(&cam_config, 4 + sizeof(cam_config.data.format7Settings.imageSettings));
    if (ret_code == RET_SUCCESS)
    {
    	m_camera_state.format7PacketSize =cam_config.data.format7Settings.packetInfo.recommendedBytesPerPacket;
    }
    else
    {
    	LOG("FAILED validating format 7 info\n");
    }

    LOG_DEBUG("Sending configuration packet with size %d\n", m_camera_state.format7PacketSize);
	cam_config.type = CAMERA_CONFIG_fc2SetFormat7Configuration;
	memcpy(&cam_config.data.format7Configuration.imageSettings,
			&m_camera_state.format7ImageSettings,
				sizeof(m_camera_state.format7ImageSettings));
	cam_config.data.format7Configuration.percentage = 100;
	request.request.p_payload = (void*)&cam_config;
	request.request.payload_len = 4 + sizeof(m_camera_state.format7ImageSettings) + sizeof(m_camera_state.format7PacketSize);
	ret_code = issue_request(&request, &response);

	if (ret_code != RET_SUCCESS)
	{
		LOG("Failed issueing request, %d\n", ret_code);
		return 0;
	}
	if (response.response.ret_code != RET_SUCCESS)
	{
		LOG("Failed setting format 7 configuration packet request, %d\n", response.response.ret_code);
		return 0;
	}

	request.request.p_payload = (void*)&cam_config;
	LOG_DEBUG("Calling fc2SetProperty for frame rate\n");
	cam_config.type = CAMERA_CONFIG_fc2SetProperty;
	memcpy(&cam_config.data.property, &m_camera_state.properties[FC2_FRAME_RATE], sizeof(m_camera_state.properties[FC2_FRAME_RATE]));
	request.request.payload_len = 4 + sizeof(m_camera_state.properties[FC2_FRAME_RATE]);
	ret_code = issue_request(&request, &response);
	RETURN_ON_ERROR(ret_code);
	RETURN_ON_ERROR(response.response.ret_code);

	request.request.p_payload = (void*)&cam_config;
	LOG_DEBUG("Calling fc2SetProperty for auto exposure rate\n");
	cam_config.type = CAMERA_CONFIG_fc2SetProperty;
	memcpy(&cam_config.data.property, &m_camera_state.properties[FC2_AUTO_EXPOSURE], sizeof(m_camera_state.properties[FC2_FRAME_RATE]));
	request.request.payload_len = 4 + sizeof(m_camera_state.properties[FC2_FRAME_RATE]);
	ret_code = issue_request(&request, &response);
	RETURN_ON_ERROR(ret_code);
	RETURN_ON_ERROR(response.response.ret_code);

	request.request.p_payload = (void*)&cam_config;
	LOG_DEBUG("Calling fc2SetProperty for shutter rate\n");
	cam_config.type = CAMERA_CONFIG_fc2SetProperty;
	memcpy(&cam_config.data.property, &m_camera_state.properties[FC2_SHUTTER], sizeof(m_camera_state.properties[FC2_FRAME_RATE]));
	request.request.payload_len = 4 + sizeof(m_camera_state.properties[FC2_FRAME_RATE]);
	ret_code = issue_request(&request, &response);
	RETURN_ON_ERROR(ret_code);
	RETURN_ON_ERROR(response.response.ret_code);

	request.request.p_payload = (void*)&cam_config;
	LOG_DEBUG("Calling fc2SetProperty for gain\n");
	cam_config.type = CAMERA_CONFIG_fc2SetProperty;
	memcpy(&cam_config.data.property, &m_camera_state.properties[FC2_GAIN], sizeof(m_camera_state.properties[FC2_GAIN]));
	request.request.payload_len = 4 + sizeof(m_camera_state.properties[FC2_GAIN]);
	ret_code = issue_request(&request, &response);
	RETURN_ON_ERROR(ret_code);
	RETURN_ON_ERROR(response.response.ret_code);

	request.request.p_payload = (void*)&cam_config;
	LOG_DEBUG("Calling fc2SetProperty for trigger delay\n");
	cam_config.type = CAMERA_CONFIG_fc2SetProperty;
	memcpy(&cam_config.data.property, &m_camera_state.properties[FC2_TRIGGER_DELAY], sizeof(m_camera_state.properties[FC2_TRIGGER_DELAY]));
	request.request.payload_len = 4 + sizeof(m_camera_state.properties[FC2_TRIGGER_DELAY]);
	ret_code = issue_request(&request, &response);
	RETURN_ON_ERROR(ret_code);
	RETURN_ON_ERROR(response.response.ret_code);


	LOG_DEBUG("Setting embedded image info\n");
	cam_config.type = CAMERA_CONFIG_fc2SetEmbeddedImageInfo;
	memcpy(&cam_config.data.embeddedImageInfo, &m_camera_state.embeddedImageInfo, sizeof(m_camera_state.embeddedImageInfo));
	request.request.p_payload = (void*)&cam_config;
	request.request.payload_len = 4 + sizeof(m_camera_state.embeddedImageInfo);
	ret_code = issue_request(&request, &response);
	RETURN_ON_ERROR(ret_code);
	RETURN_ON_ERROR(response.response.ret_code);

	LOG_DEBUG("Setting trigger mode\n");
	cam_config.type = CAMERA_CONFIG_fc2SetTriggerMode;
	memcpy(&cam_config.data.triggerMode, &m_camera_state.triggerMode, sizeof(m_camera_state.triggerMode));
	request.request.p_payload = (void*)&cam_config;
	request.request.payload_len = 4 + sizeof(m_camera_state.triggerMode);
	ret_code = issue_request(&request, &response);
	RETURN_ON_ERROR(ret_code);
	RETURN_ON_ERROR(response.response.ret_code);


	return RET_SUCCESS;
}

ret_code_t camera_config(camera_config_t * p_config, size_t config_size)
{
	ret_code_t ret_code;

	message_t request, response;
	request.type = MSG_TYPE_CONFIG_CAMERA;
	request.request.p_payload = (void*)p_config;
	request.request.payload_len = config_size;

	ret_code = issue_request(&request, &response);
	RETURN_ON_ERROR(ret_code);
	RETURN_ON_ERROR(response.response.ret_code);

	memcpy(p_config, response.response.p_payload, response.response.payload_len);

	return RET_SUCCESS;
}

void print_config_data(void)
{
	ret_code_t ret_code;
    camera_config_t config;

    config.type = CAMERA_CONFIG_fc2GetConfiguration;
    ret_code = camera_config(&config, 4);
    if (ret_code == RET_SUCCESS)
    {
    	LOG("\nGot camera config\n");

    	LOG("asyncBusSpeed is %d\n", config.data.config.asyncBusSpeed);
    	LOG("bandwidthAllocation is %d\n", config.data.config.bandwidthAllocation);
    	LOG("grabMode is %d\n", config.data.config.grabMode);
    	LOG("grabTimeout is %d\n", config.data.config.grabTimeout);
    	LOG("isochBusSpeed is %d\n", config.data.config.isochBusSpeed);
    	LOG("minNumImageNotifications is %d\n", config.data.config.minNumImageNotifications);
    	LOG("numBuffers is %d\n", config.data.config.numBuffers);
    	LOG("numImageNotifications is %d\n", config.data.config.numImageNotifications);
    	LOG("registerTimeout is %d\n", config.data.config.registerTimeout);
    	LOG("registerTimeoutRetries is %d\n", config.data.config.registerTimeoutRetries);
    }
    else
    {
    	LOG("FAILED getting config\n");
    }

    config.type = CAMERA_CONFIG_fc2GetVideoModeAndFrameRate;
    ret_code = camera_config(&config, 4);
    if (ret_code == RET_SUCCESS)
    {
    	LOG("\nGot video and framerate\n");

    	LOG("videoMode is %d\n", config.data.video_framerate_info.videoMode);
    	LOG("frameRate is %d\n", config.data.video_framerate_info.frameRate);
    }
    else
    {
    	LOG("FAILED getting video and frame rate\n");
    }

    config.type = CAMERA_CONFIG_fc2GetFormat7Info;
    config.data.format7Info.format7Info.mode = FC2_MODE_0;
    ret_code = camera_config(&config, 4 + sizeof(config.data.format7Info.format7Info.mode));
    if (ret_code == RET_SUCCESS)
    {
    	LOG("\nGot format 7 info\n");

    	LOG("mode is %d\n", config.data.format7Info.format7Info.mode);
    	LOG("supported is %d\n", config.data.format7Info.supported);
    	LOG("imageHStepSize is %d\n", config.data.format7Info.format7Info.imageHStepSize);
    	LOG("imageVStepSize is %d\n", config.data.format7Info.format7Info.imageVStepSize);
    	LOG("maxHeight is %d\n", config.data.format7Info.format7Info.maxHeight);
    	LOG("maxWidth is %d\n", config.data.format7Info.format7Info.maxWidth);
    	LOG("maxPacketSize is %d\n", config.data.format7Info.format7Info.maxPacketSize);
    	LOG("minPacketSize is %d\n", config.data.format7Info.format7Info.minPacketSize);
    	LOG("offsetHStepSize is %d\n", config.data.format7Info.format7Info.offsetHStepSize);
    	LOG("offsetVStepSize is %d\n", config.data.format7Info.format7Info.offsetVStepSize);
    	LOG("packetSize is %d\n", config.data.format7Info.format7Info.packetSize);
    	LOG("percentage is %f\n", config.data.format7Info.format7Info.percentage);
    	LOG("pixelFormatBitField is 0x%x\n", config.data.format7Info.format7Info.pixelFormatBitField);
    	LOG("vendorPixelFormatBitField is %u\n", config.data.format7Info.format7Info.vendorPixelFormatBitField);
    }
    else
    {
    	LOG("FAILED getting format 7 info\n");
    }

    config.type = CAMERA_CONFIG_fc2GetFormat7Configuration;
    ret_code = camera_config(&config, 4);
    if (ret_code == RET_SUCCESS)
    {
    	LOG("\nGot format 7 configuration\n");

    	LOG("mode is %d\n", config.data.format7Configuration.imageSettings.mode);
    	LOG("height is %d\n", config.data.format7Configuration.imageSettings.height);
    	LOG("width is %d\n", config.data.format7Configuration.imageSettings.width);
    	LOG("offsetX is %d\n", config.data.format7Configuration.imageSettings.offsetX);
    	LOG("offsetY is %d\n", config.data.format7Configuration.imageSettings.offsetY);
    	LOG("pixelFormat is 0x%x\n", config.data.format7Configuration.imageSettings.pixelFormat);
    	LOG("packetSize is %d\n", config.data.format7Configuration.packetSize);
    	LOG("percentage is %f\n", config.data.format7Configuration.percentage);
    }
    else
    {
    	LOG("FAILED getting format 7 configuration\n");
    }


    config.type = CAMERA_CONFIG_fc2ValidateFormat7Settings;
    config.data.format7Settings.imageSettings.height = 600;
    config.data.format7Settings.imageSettings.width  = 1248;
    config.data.format7Settings.imageSettings.offsetX = 400;
    config.data.format7Settings.imageSettings.offsetY = 400;
    config.data.format7Settings.imageSettings.pixelFormat = FC2_PIXEL_FORMAT_MONO8;
    config.data.format7Settings.imageSettings.mode = FC2_MODE_0;

    ret_code = camera_config(&config, 4 + sizeof(config.data.format7Settings.imageSettings));
    if (ret_code == RET_SUCCESS)
    {
    	LOG("\nGot format7 validation\n");

    	LOG("Supported is %d\n", config.data.format7Settings.supported);
    	LOG("maxBytesPerPacket is %d\n", config.data.format7Settings.packetInfo.maxBytesPerPacket);
    	LOG("recommendedBytesPerPacket is %d\n", config.data.format7Settings.packetInfo.recommendedBytesPerPacket);
    	LOG("unitBytesPerPacket is %d\n", config.data.format7Settings.packetInfo.unitBytesPerPacket);
    }
    else
    {
    	LOG("FAILED validating format 7 info\n");
    }

    config.type = CAMERA_CONFIG_fc2GetPropertyInfo;
    config.data.propertyInfo.type = FC2_FRAME_RATE;
    ret_code = camera_config(&config, 4 + sizeof(config.data.propertyInfo.type));
    if (ret_code == RET_SUCCESS)
    {
    	LOG("\nGot framerate property info\n");

    	LOG("absMax is %f\n", config.data.propertyInfo.absMax);
    	LOG("absMin is %f\n", config.data.propertyInfo.absMin);
    	LOG("absValSupported is %d\n", config.data.propertyInfo.absValSupported);
    	LOG("autoSupported is %d\n", config.data.propertyInfo.autoSupported);
    	LOG("manualSupported is %d\n", config.data.propertyInfo.manualSupported);
    	LOG("max is %d\n", config.data.propertyInfo.max);
    	LOG("min is %d\n", config.data.propertyInfo.min);
    	LOG("onOffSupported is %d\n", config.data.propertyInfo.onOffSupported);
    	LOG("onePushSupported is %d\n", config.data.propertyInfo.onePushSupported);
    	LOG("pUnitAbbr is %s\n", config.data.propertyInfo.pUnitAbbr);
    	LOG("pUnits is %s\n", config.data.propertyInfo.pUnits);
    	LOG("present is %d\n", config.data.propertyInfo.present);
    	LOG("readOutSupported is %d\n", config.data.propertyInfo.readOutSupported);
    }
    else
    {
    	LOG("FAILED getting framerate property info\n");
    }


    config.type = CAMERA_CONFIG_fc2GetProperty;
    config.data.property.type = FC2_FRAME_RATE;
    ret_code = camera_config(&config, 4 + sizeof(config.data.property.type));
    if (ret_code == RET_SUCCESS)
    {
    	LOG("\nGot framerate property\n");

    	LOG("present is %d\n", config.data.property.present);
    	LOG("absControl is %d\n", config.data.property.absControl);
    	LOG("absValue is %f\n", config.data.property.absValue);
    	LOG("autoManualMode is %d\n", config.data.property.autoManualMode);
    	LOG("onOff is %d\n", config.data.property.onOff);
    	LOG("onePush is %d\n", config.data.property.onePush);
    	LOG("valueA is %d\n", config.data.property.valueA);
    	LOG("valueB is %d\n", config.data.property.valueB);
    }
    else
    {
    	LOG("FAILED getting framerate property\n");
    }

    config.type = CAMERA_CONFIG_fc2GetPropertyInfo;
    config.data.propertyInfo.type = FC2_SHUTTER;
    ret_code = camera_config(&config, 4 + sizeof(config.data.propertyInfo.type));
    if (ret_code == RET_SUCCESS)
    {
    	LOG("\nGot shutter property info\n");

    	LOG("absMax is %f\n", config.data.propertyInfo.absMax);
    	LOG("absMin is %f\n", config.data.propertyInfo.absMin);
    	LOG("absValSupported is %d\n", config.data.propertyInfo.absValSupported);
    	LOG("autoSupported is %d\n", config.data.propertyInfo.autoSupported);
    	LOG("manualSupported is %d\n", config.data.propertyInfo.manualSupported);
    	LOG("max is %d\n", config.data.propertyInfo.max);
    	LOG("min is %d\n", config.data.propertyInfo.min);
    	LOG("onOffSupported is %d\n", config.data.propertyInfo.onOffSupported);
    	LOG("onePushSupported is %d\n", config.data.propertyInfo.onePushSupported);
    	LOG("pUnitAbbr is %s\n", config.data.propertyInfo.pUnitAbbr);
    	LOG("pUnits is %s\n", config.data.propertyInfo.pUnits);
    	LOG("present is %d\n", config.data.propertyInfo.present);
    	LOG("readOutSupported is %d\n", config.data.propertyInfo.readOutSupported);
    }
    else
    {
    	LOG("FAILED getting shutter property info\n");
    }

    config.type = CAMERA_CONFIG_fc2GetProperty;
    config.data.property.type = FC2_SHUTTER;
    ret_code = camera_config(&config, 4 + sizeof(config.data.property.type));
    if (ret_code == RET_SUCCESS)
    {
    	LOG("\nGot shutter property\n");

    	LOG("present is %d\n", config.data.property.present);
    	LOG("absControl is %d\n", config.data.property.absControl);
    	LOG("absValue is %f\n", config.data.property.absValue);
    	LOG("autoManualMode is %d\n", config.data.property.autoManualMode);
    	LOG("onOff is %d\n", config.data.property.onOff);
    	LOG("onePush is %d\n", config.data.property.onePush);
    	LOG("valueA is %d\n", config.data.property.valueA);
    	LOG("valueB is %d\n", config.data.property.valueB);
    }
    else
    {
    	LOG("FAILED getting shutter property\n");
    }

    config.type = CAMERA_CONFIG_fc2GetPropertyInfo;
    config.data.propertyInfo.type = FC2_AUTO_EXPOSURE;
    ret_code = camera_config(&config, 4 + sizeof(config.data.propertyInfo.type));
    if (ret_code == RET_SUCCESS)
    {
    	LOG("\nGot auto exposure property info\n");

    	LOG("absMax is %f\n", config.data.propertyInfo.absMax);
    	LOG("absMin is %f\n", config.data.propertyInfo.absMin);
    	LOG("absValSupported is %d\n", config.data.propertyInfo.absValSupported);
    	LOG("autoSupported is %d\n", config.data.propertyInfo.autoSupported);
    	LOG("manualSupported is %d\n", config.data.propertyInfo.manualSupported);
    	LOG("max is %d\n", config.data.propertyInfo.max);
    	LOG("min is %d\n", config.data.propertyInfo.min);
    	LOG("onOffSupported is %d\n", config.data.propertyInfo.onOffSupported);
    	LOG("onePushSupported is %d\n", config.data.propertyInfo.onePushSupported);
    	LOG("pUnitAbbr is %s\n", config.data.propertyInfo.pUnitAbbr);
    	LOG("pUnits is %s\n", config.data.propertyInfo.pUnits);
    	LOG("present is %d\n", config.data.propertyInfo.present);
    	LOG("readOutSupported is %d\n", config.data.propertyInfo.readOutSupported);
    }
    else
    {
    	LOG("FAILED getting exposure property info\n");
    }

    config.type = CAMERA_CONFIG_fc2GetProperty;
    config.data.property.type = FC2_AUTO_EXPOSURE;
    ret_code = camera_config(&config, 4 + sizeof(config.data.property.type));
    if (ret_code == RET_SUCCESS)
    {
    	LOG("\nGot auto exposure property\n");

    	LOG("present is %d\n", config.data.property.present);
    	LOG("absControl is %d\n", config.data.property.absControl);
    	LOG("absValue is %f\n", config.data.property.absValue);
    	LOG("autoManualMode is %d\n", config.data.property.autoManualMode);
    	LOG("onOff is %d\n", config.data.property.onOff);
    	LOG("onePush is %d\n", config.data.property.onePush);
    	LOG("valueA is %d\n", config.data.property.valueA);
    	LOG("valueB is %d\n", config.data.property.valueB);
    }
    else
    {
    	LOG("FAILED getting auto exposure property\n");
    }

    config.type = CAMERA_CONFIG_fc2GetPropertyInfo;
    config.data.propertyInfo.type = FC2_GAIN;
    ret_code = camera_config(&config, 4 + sizeof(config.data.propertyInfo.type));
    if (ret_code == RET_SUCCESS)
    {
    	LOG("\nGot gain property info\n");

    	LOG("absMax is %f\n", config.data.propertyInfo.absMax);
    	LOG("absMin is %f\n", config.data.propertyInfo.absMin);
    	LOG("absValSupported is %d\n", config.data.propertyInfo.absValSupported);
    	LOG("autoSupported is %d\n", config.data.propertyInfo.autoSupported);
    	LOG("manualSupported is %d\n", config.data.propertyInfo.manualSupported);
    	LOG("max is %d\n", config.data.propertyInfo.max);
    	LOG("min is %d\n", config.data.propertyInfo.min);
    	LOG("onOffSupported is %d\n", config.data.propertyInfo.onOffSupported);
    	LOG("onePushSupported is %d\n", config.data.propertyInfo.onePushSupported);
    	LOG("pUnitAbbr is %s\n", config.data.propertyInfo.pUnitAbbr);
    	LOG("pUnits is %s\n", config.data.propertyInfo.pUnits);
    	LOG("present is %d\n", config.data.propertyInfo.present);
    	LOG("readOutSupported is %d\n", config.data.propertyInfo.readOutSupported);
    }
    else
    {
    	LOG("FAILED getting gain property info\n");
    }

    config.type = CAMERA_CONFIG_fc2GetProperty;
	config.data.property.type = FC2_GAIN;
	ret_code = camera_config(&config, 4 + sizeof(config.data.property.type));
	if (ret_code == RET_SUCCESS)
	{
		LOG("\nGot gain property\n");

		LOG("present is %d\n", config.data.property.present);
		LOG("absControl is %d\n", config.data.property.absControl);
		LOG("absValue is %f\n", config.data.property.absValue);
		LOG("autoManualMode is %d\n", config.data.property.autoManualMode);
		LOG("onOff is %d\n", config.data.property.onOff);
		LOG("onePush is %d\n", config.data.property.onePush);
		LOG("valueA is %d\n", config.data.property.valueA);
		LOG("valueB is %d\n", config.data.property.valueB);
	}
	else
	{
		LOG("FAILED getting gain property\n");
	}

    config.type = CAMERA_CONFIG_fc2GetEmbeddedImageInfo;
    ret_code = camera_config(&config, 4);
    if (ret_code == RET_SUCCESS)
    {
    	fc2EmbeddedImageInfo * p_info = &config.data.embeddedImageInfo;
    	LOG("\nGot embedded image info\n");


    	LOG("timestamp;      %d, %d\n", p_info->timestamp.available, p_info->timestamp.onOff);
		LOG("gain;           %d, %d\n", p_info->gain.available, p_info->gain.onOff);
		LOG("shutter;        %d, %d\n", p_info->shutter.available, p_info->shutter.onOff);
		LOG("brightness;     %d, %d\n", p_info->brightness.available, p_info->brightness.onOff);
		LOG("exposure;       %d, %d\n", p_info->exposure.available, p_info->exposure.onOff);
		LOG("whiteBalance;   %d, %d\n", p_info->whiteBalance.available, p_info->whiteBalance.onOff);
		LOG("frameCounter;   %d, %d\n", p_info->frameCounter.available, p_info->frameCounter.onOff);
		LOG("strobePattern;  %d, %d\n", p_info->strobePattern.available, p_info->strobePattern.onOff);
		LOG("GPIOPinState;   %d, %d\n", p_info->GPIOPinState.available, p_info->GPIOPinState.onOff);
		LOG("ROIPosition;    %d, %d\n", p_info->ROIPosition.available, p_info->ROIPosition.onOff);

    }
    else
    {
    	LOG("Failed getting embedded image info\n");
    }

    config.type = CAMERA_CONFIG_fc2GetTriggerModeInfo;
    ret_code = camera_config(&config, 4);
    if (ret_code == RET_SUCCESS)
    {
    	fc2TriggerModeInfo * p_info = &config.data.triggerModeInfo;
    	LOG("\nGot trigger mode info\n");

    	LOG("present;                    %d\n", p_info->present);
		LOG("readOutSupported;           %d\n", p_info->readOutSupported);
		LOG("onOffSupported;             %d\n", p_info->onOffSupported);
		LOG("polaritySupported;          %d\n", p_info->polaritySupported);
		LOG("valueReadable;              %d\n", p_info->valueReadable);
		LOG("sourceMask;                 %d\n", p_info->sourceMask);
		LOG("softwareTriggerSupported;   %d\n", p_info->softwareTriggerSupported);
		LOG("modeMask;                   %d\n", p_info->modeMask);
    }
    else
    {
    	LOG("Failed getting trigger mode info\n");
    }

    config.type = CAMERA_CONFIG_fc2GetTriggerMode;
    ret_code = camera_config(&config, 4);
    if (ret_code == RET_SUCCESS)
    {
    	fc2TriggerMode * p_info = &config.data.triggerMode;
    	LOG("\nGot trigger mode\n");

    	LOG("onOff;         %d\n", p_info->onOff);
		LOG("polarity;      %d\n", p_info->polarity);
		LOG("source;        %d\n", p_info->source);
		LOG("mode;          %d\n", p_info->mode);
		LOG("parameter;     %d\n", p_info->parameter);
    }
    else
    {
    	LOG("Failed getting trigger mode\n");
    }

    config.type = CAMERA_CONFIG_fc2GetTriggerDelayInfo;
    ret_code = camera_config(&config, 4);
    if (ret_code == RET_SUCCESS)
    {
    	LOG("\nGot Trigger delay info\n");

    	LOG("absMax is %f\n", config.data.triggerDelayInfo.absMax);
    	LOG("absMin is %f\n", config.data.triggerDelayInfo.absMin);
    	LOG("absValSupported is %d\n", config.data.triggerDelayInfo.absValSupported);
    	LOG("autoSupported is %d\n", config.data.triggerDelayInfo.autoSupported);
    	LOG("manualSupported is %d\n", config.data.triggerDelayInfo.manualSupported);
    	LOG("max is %d\n", config.data.triggerDelayInfo.max);
    	LOG("min is %d\n", config.data.triggerDelayInfo.min);
    	LOG("onOffSupported is %d\n", config.data.triggerDelayInfo.onOffSupported);
    	LOG("onePushSupported is %d\n", config.data.triggerDelayInfo.onePushSupported);
    	LOG("pUnitAbbr is %s\n", config.data.triggerDelayInfo.pUnitAbbr);
    	LOG("pUnits is %s\n", config.data.triggerDelayInfo.pUnits);
    	LOG("present is %d\n", config.data.triggerDelayInfo.present);
    	LOG("readOutSupported is %d\n", config.data.triggerDelayInfo.readOutSupported);
    }
    else
    {
    	LOG("FAILED getting trigger delay info\n");
    }

    config.type = CAMERA_CONFIG_fc2GetTriggerDelay;
    ret_code = camera_config(&config, 4);
    if (ret_code == RET_SUCCESS)
    {
    	LOG("\nGot Trigger delay info\n");

    	LOG("present is %d\n", config.data.triggerDelay.present);
    	LOG("absControl is %d\n", config.data.triggerDelay.absControl);
    	LOG("absValue is %f\n", config.data.triggerDelay.absValue);
    	LOG("autoManualMode is %d\n", config.data.triggerDelay.autoManualMode);
    	LOG("onOff is %d\n", config.data.triggerDelay.onOff);
    	LOG("onePush is %d\n", config.data.triggerDelay.onePush);
    	LOG("valueA is %d\n", config.data.triggerDelay.valueA);
    	LOG("valueB is %d\n", config.data.triggerDelay.valueB);
    }
    else
    {
    	LOG("FAILED getting trigger delay info\n");
    }



}
