#ifndef FRAME_BUFFER_H__
#define FRAME_BUFFER_H__

#ifdef __cplusplus
extern "C"
{
#endif


#include <stdint.h>
#include "ret_code.h"

#if !defined(FRAME_BUFFER_STORE_DEVICE) && \
    !defined(FRAME_BUFFER_STORE_HOST)
#error "Specify frame buffer storage configuration"
#endif

#if !defined(FRAME_BUFFER_COPY_ASYNC) && \
    !defined(FRAME_BUFFER_COPY_SYNC)
#error "Specify frame buffer copy configuration"
#endif


/** @brief This struct holds the state of a frame buffer. */
typedef struct frame_buffer_s
{
    uint8_t  * d_image_buffer;      /**< Pointer to frame buffer memory. Images are stored in either host or device memory. */
    uint32_t num_images;            /**< The current number of images stored in the buffer. */
    uint32_t max_images;            /**< The image buffer size in images. */

    uint32_t s_width;               /**< The width of images in pixels. */
    uint32_t s_height;              /**< The height of the onboard images in pixels. */

    uint32_t image_size;            /**< Size per image in bytes. */
    cudaStream_t * p_stream;
} frame_buffer_t;


/** @brief Initialization structure needed for frame buffer initialization. */
typedef struct
{
    uint32_t image_height;      /**< Height of the images in pixels. */
    uint32_t image_width;       /**< Width of the images in bytes. */
    uint32_t max_images;        /**< The size of the image buffer in images. */
    cudaStream_t * p_stream;    /**< A pointer to the assigned CUDA stream used for memory transfers. */
} frame_buffer_init_t;


/** @brief Initializes a frame buffer
 *
 *  @details Allocates the necessary memory defined by the given configuration.
 *
 *  @param[inout] p_frame_buffer        Pointer to the structure holding the state of the frame buffer.
 *  @param[in]  p_init                  Pointer to frame buffer configuration.
 *
 *  @return RET_SUCCESS                 If the frame buffer was successfully initialized.
 *  @return RET_ERROR_NULL              If a NULL parameter was provided.
 *  @return RET_ERROR_INVALID_CONFIG    If a invalid configuration was given.
 *  @return RET_ERROR_NO_MEM            If the system failed to allocate enough memory for the frame buffers.
 */
ret_code_t frame_buffer_init(frame_buffer_t * p_frame_buffer, frame_buffer_init_t * p_init);


/** @brief Deinitializes a frame buffer
 *
 *  @details Deallocates the frame buffer memory. Cleans up state variables.
 *
 *  @param[inout] p_frame_buffer        Pointer to the structure holding the state of the frame buffer.
 *
 *  @return RET_SUCCESS                 If the frame buffer was successfully deinitialized.
 *  @return RET_ERROR_NULL              If a NULL parameter was provided.
 */
ret_code_t frame_buffer_deinit(frame_buffer_t * frame_buffer);


/** @brief Adds an image to a frame buffer
 *
 *  @details The image is expected to have the same dimension as the frame buffer configuration.
 *           The image is copied to either host or device memory, depending on the compile time flags
 *           FRAME_BUFFER_STORE_DEVICE and FRAME_BUFFER_STORE_HOST.
 *
 *  @param[inout] p_frame_buffer        Pointer to the structure holding the state of the frame buffer.
 *  @param[in]    p_image               Pointer to the the image memory to be stored in the frame buffer.
 *
 *  @return RET_SUCCESS                 If the image was successfully added to the frame buffer.
 *  @return RET_ERROR_NULL              If a NULL parameter was provided.
 *  @return RET_ERROR_NO_MEM            If there is no available memory in the frame buffer.
 *  @return RET_ERROR_OTHER             If an error during image copy.
 */
ret_code_t frame_buffer_add_image(frame_buffer_t * p_frame_buffer, uint8_t * p_image);


/** @brief Removes all images from the frame buffer.
 *
 *  @details For performance reasons, the memory is not cleared. However, the image counter is set to 0.
 *
 *  @param[inout] p_frame_buffer        Pointer to the structure holding the state of the frame buffer.
 *
 *  @return RET_SUCCESS                 If the frame buffer was successfully cleared.
 *  @return RET_ERROR_NULL              If a NULL parameter was provided.
 */
ret_code_t frame_buffer_clear(frame_buffer_t * p_frame_buffer);


#ifdef __cplusplus
}
#endif

#endif // FRAME_BUFFER_H__
