#ifndef GPIO_H__
#define GPIO_H__

#ifdef __cplusplus
extern "C"
{
#endif


/* This implementation is inspired by https://github.com/jetsonhacks/jetsonGPIO.
 * This library does only provide simple GPIO function needed for this project.
 */


#include "ret_code.h"
#include <stdbool.h>


/** @brief This enum defines the GPIO pin values for the Jetson TK1 and TX1. */
typedef enum
{
    // Jetson TK1 pins
    gpio57  =  57,    // J3A1 - Pin 50
    gpio160 = 160,    // J3A2 - Pin 40
    gpio161 = 161,    // J3A2 - Pin 43
    gpio162 = 162,    // J3A2 - Pin 46
    gpio163 = 163,    // J3A2 - Pin 49
    gpio164 = 164,    // J3A2 - Pin 52
    gpio165 = 165,    // J3A2 - Pin 55
    gpio166 = 166,    // J3A2 - Pin 58

    // Jetson TX1 pins
    gpio36 = 36,      // J21 - Pin 32 - Unused - AO_DMIC_IN_CLK
    gpio37 = 37,      // J21 - Pin 16 - Unused - AO_DMIC_IN_DAT
    gpio38 = 38,      // J21 - Pin 13 - Bidir  - GPIO20/AUD_INT
    gpio63 = 63,      // J21 - Pin 33 - Bidir  - GPIO11_AP_WAKE_BT
    gpio184 = 184,    // J21 - Pin 18 - Input  - GPIO16_MDM_WAKE_AP
    gpio186 = 186,    // J21 - Pin 31 - Input  - GPIO9_MOTION_INT
    gpio187 = 187,    // J21 - Pin 37 - Output - GPIO8_ALS_PROX_INT
    gpio219 = 219     // J21 - Pin 29 - Output - GPIO19_AUD_RST
} gpio_pin_t;


/** @brief Initializes a GPIO pin for output
 *
 *  @details Exports the GPIO file descriptor and sets the direction of the pin.
 *
 *  @param[in] pin                      The pin used for GPIO initialization.
 *
 *  @return RET_SUCCESS                 If the pin was successfully configured for output.
 *  @return RET_ERROR_OTHER             If an error occurred.
 */
ret_code_t gpio_init_output(gpio_pin_t pin);


/** @brief Deinitializes a GPIO pin for output
 *
 *  @details Unexports the GPIO file descriptor.
 *
 *  @param[in] pin                      The pin used for GPIO deinitialization.
 *
 *  @return RET_SUCCESS                 If the pin was successfully configured for output.
 *  @return RET_ERROR_OTHER             If an error occurred.
 */
ret_code_t gpio_deinit_output(gpio_pin_t pin);


/** @brief Sets the value of a GPIO pin
 *
 *  @details This operation can only be performed when the GPIO pin is configured for output.
 *
 *  @param[in] pin                      The pin used for GPIO initialization.
 *  @param[in] high                     If this value is set to true, the pin value will be set to 1.8V or 3.2V, representing logical high.
 *
 *  @return RET_SUCCESS                 If the pin value was successfully set.
 *  @return RET_ERROR_INVALID_STATE     If the pin was not properly configured.
 *  @return RET_ERROR_OTHER             If an error occurred.
 */
ret_code_t gpio_pin_set(gpio_pin_t pin, bool high);

#ifdef __cplusplus
}
#endif



#endif
