#include "img_processing.h"

#include <stdbool.h>
#include "scanner.h"
#include "app_error.h"
#include "frame_buffer.h"
#include "tcp_server.h"

static bool m_image_processing_initialized;


typedef void (*img_process_kernel)(frame_buffer_t * d_frame_buffers,
                                   scan_output_buffer_t * d_output_buffer);

static img_process_kernel      m_img_process_kernel_func;
static img_processing_config_t m_img_process_config;

// forward declarations of the image processing init functions
ret_code_t scan_mode_0_init(scan_config_t * p_config);
ret_code_t scan_mode_1_init(scan_config_t * p_config);
ret_code_t scan_mode_2_init(scan_config_t * p_config);
ret_code_t scan_mode_3_init(scan_config_t * p_config);
ret_code_t scan_mode_4_init(scan_config_t * p_config);


// forward declarations of the image processing kernels
__global__ void scan_mode_0_kernel(frame_buffer_t * p_frame_buffers, scan_output_buffer_t * d_output_buffer);
__global__ void scan_mode_1_kernel(frame_buffer_t * p_frame_buffers, scan_output_buffer_t * d_output_buffer);
__global__ void scan_mode_2_kernel(frame_buffer_t * p_frame_buffers, scan_output_buffer_t * d_output_buffer);
__global__ void scan_mode_3_kernel(frame_buffer_t * p_frame_buffers, scan_output_buffer_t * d_output_buffer);
__global__ void scan_mode_4_kernel(frame_buffer_t * p_frame_buffers, scan_output_buffer_t * d_output_buffer);


static ret_code_t validate_config(scan_config_t * p_scan_config)
{
    switch (p_scan_config->scan_mode)
    {
        case SCAN_MODE_0: break;
        case SCAN_MODE_1: break;
        case SCAN_MODE_2: break;
        case SCAN_MODE_3: break;
        case SCAN_MODE_4: break;
        default:          return RET_ERROR_NOT_IMPLEMENTED;
    }

    if (p_scan_config->img_processing_config.frames_per_message == 0)
    {
        return RET_ERROR_INVALID_CONFIG;
    }
    if (p_scan_config->img_processing_config.threads_per_block == 0)
    {
        return RET_ERROR_INVALID_CONFIG;
    }

    if (p_scan_config->img_processing_config.scatter_distance % 2 != 0)
    {
        LOG("Scatter distance must be a power of 2\n");
        return RET_ERROR_INVALID_CONFIG;
    }

    return RET_SUCCESS;
}



ret_code_t img_processing_set_config(scan_config_t * p_scan_config)
{
    if (p_scan_config == NULL)
    {
        return RET_ERROR_NULL;
    }

    ret_code_t ret_code = RET_SUCCESS;

    ret_code = validate_config(p_scan_config);
    RETURN_ON_ERROR(ret_code);

    // set function pointer
    switch (p_scan_config->scan_mode)
    {
        case SCAN_MODE_0:
            m_img_process_kernel_func = scan_mode_0_kernel; ret_code = scan_mode_0_init(p_scan_config); break;
        case SCAN_MODE_1:
            m_img_process_kernel_func = scan_mode_1_kernel; ret_code = scan_mode_1_init(p_scan_config); break;
        case SCAN_MODE_2:
            m_img_process_kernel_func = scan_mode_2_kernel; ret_code = scan_mode_2_init(p_scan_config); break;
        case SCAN_MODE_3:
            m_img_process_kernel_func = scan_mode_3_kernel; ret_code = scan_mode_3_init(p_scan_config); break;
        case SCAN_MODE_4:
            m_img_process_kernel_func = scan_mode_4_kernel; ret_code = scan_mode_4_init(p_scan_config); break;
        default:
            ret_code =  RET_ERROR_INVALID_CONFIG;
            break;
    }

    if (ret_code != RET_SUCCESS)
    {
        return ret_code;
    }

    memcpy(&m_img_process_config, &p_scan_config->img_processing_config, sizeof(img_processing_config_t));

    m_image_processing_initialized = true;

    return RET_SUCCESS;
}


static void on_image_processing_completed(cudaStream_t stream, cudaError_t status, void *user_data)
{

    stream_data_t * p_stream_data = (stream_data_t *)user_data;

    *p_stream_data->h_scan_output_buffer->p_gpu_execution_time   = p_stream_data->prev_gpu_processing_time;
    *p_stream_data->h_scan_output_buffer->p_total_execution_time = p_stream_data->prev_total_processing_time;

    const message_t msg = {.type                = MSG_TYPE_SCAN_DATA,
                           {
                            .request = {.p_payload   = p_stream_data->h_scan_output_buffer->p_buffer,
                                        .payload_len = p_stream_data->h_scan_output_buffer->buffer_size
                                        }
                           }
                          };
    tcp_server_send_message(&msg);
}


ret_code_t img_processing_start(stream_data_t  * p_stream_data)
{
    if (p_stream_data == NULL)
    {
        return RET_ERROR_NULL;
    }

    if (!m_image_processing_initialized)
    {
        return RET_ERROR_NOT_INITIALIZED;
    }

    cudaError_t cuda_error;

    // calculate the previous execution times for this stream
    cuda_error = cudaEventElapsedTime(&p_stream_data->prev_gpu_processing_time,
                                      p_stream_data->start_event,
                                      p_stream_data->gpu_finished_event);
    if (cuda_error != cudaSuccess)
    {
        LOG_DEBUG("Failed obtaining elapsed GPU time, %s\n", cudaGetErrorString(cuda_error));
        p_stream_data->prev_gpu_processing_time = 0;
    }

    cuda_error = cudaEventElapsedTime(&p_stream_data->prev_total_processing_time,
                                      p_stream_data->start_event,
                                      p_stream_data->sending_finished_event);
    if (cuda_error != cudaSuccess)
    {
        LOG_DEBUG("Failed obtaining elapsed total time, %s\n", cudaGetErrorString(cuda_error));
        p_stream_data->prev_total_processing_time = 0;
    }

    cudaEventRecord(p_stream_data->start_event, p_stream_data->stream);

    dim3 block_dim;
    const uint32_t threads_per_block = m_img_process_config.threads_per_block;
    const uint32_t scan_lines_per_image = p_stream_data->h_scan_output_buffer->p_config->image_info.num_scan_lines_per_image;
    block_dim.x = (scan_lines_per_image + threads_per_block - 1)/ threads_per_block;
    block_dim.y = m_img_process_config.frames_per_message;
    block_dim.z = 1;

    cudaGetLastError();

    m_img_process_kernel_func<<< block_dim, threads_per_block, 0, p_stream_data->stream>>>
            (p_stream_data->d_frame_buffers, p_stream_data->d_scan_output_buffer);


    cuda_error = cudaGetLastError();
    if (cuda_error != cudaSuccess)
    {
        LOG_DEBUG("cuda failed launching kernel: %s\n", cudaGetErrorString(cuda_error));
        return RET_ERROR_OTHER;
    }

#ifdef STREAM_DATA_OUTPUT_BUFFER_DEVICE
    cuda_error = cudaMemcpyAsync(p_stream_data->h_scan_output_buffer->p_buffer,
                                 p_stream_data->d_scan_output_buffer_raw,
                                 p_stream_data->h_scan_output_buffer->buffer_size,
                                 cudaMemcpyDeviceToHost,
                                 p_stream_data->stream);
    if (cuda_error != cudaSuccess)
    {
        LOG_DEBUG("Failed copying buffer from device,  %s\n", cudaGetErrorString(cuda_error));
        return RET_ERROR_OTHER;
    }
#endif

    cudaEventRecord(p_stream_data->gpu_finished_event, p_stream_data->stream);


    cuda_error = cudaStreamAddCallback(p_stream_data->stream,
                                       on_image_processing_completed,
                                       (void*)p_stream_data,
                                       0);
    if (cuda_error != cudaSuccess)
    {
        LOG_DEBUG("cuda failed adding callback %s\n", cudaGetErrorString(cuda_error));
        return RET_ERROR_OTHER;
    }


    cudaEventRecord(p_stream_data->sending_finished_event, p_stream_data->stream);


    return RET_SUCCESS;
}




