#include <math.h>
#include "frame_buffer.h"
#include "scanner.h"
#include "log.h"

#define MAX(a,b) ((a) > (b) ? (a) : (b))
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#define GET_INDEX(column, height, offset) ((offset) + (column) +  PITCH*(height))

#define SUBTRACT(p_color_buffer, p_noise_buffer, index) \
    (p_color_buffer)[(index)] - (p_noise_buffer)[(index)]


#define LOOKUP_TABLE_LENGTH (1<<12)

static texture<float, 1, cudaReadModeElementType> lookup_table;


ret_code_t scan_mode_4_init(scan_config_t * p_config)
{
    float h_texture[LOOKUP_TABLE_LENGTH];
    cudaError_t cuda_error;
    float * d_texture;
    cudaArray *cuArray;


    for (int i = 0; i < LOOKUP_TABLE_LENGTH; i++)
    {
        // implement threshold
        if ( i > (p_config->img_processing_config.threshold_value >> 4))
        {
            h_texture[i] = pow(i,p_config->img_processing_config.pixel_power);

            // values are in the range 0 .. 2^16
            h_texture[i] = (1<<16)*(h_texture[i]/pow(LOOKUP_TABLE_LENGTH,p_config->img_processing_config.pixel_power));
        }
    }

    cuda_error = cudaMalloc( (void **) &d_texture, sizeof(h_texture));
    if (cuda_error != cudaSuccess)
    {
        LOG("Failed allocating texture memory\n");
        return RET_ERROR_NO_MEM;
    }

    // Allocate texture array and copy image data
    cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc(32, 0, 0, 0, cudaChannelFormatKindFloat);

    cuda_error = cudaMallocArray(&cuArray,
                                 &channelDesc,
                                 LOOKUP_TABLE_LENGTH * sizeof(float),
                                 1);
    if (cuda_error != cudaSuccess)
    {
        LOG("Failed allocating texture array memory\n");
        return RET_ERROR_NO_MEM;
    }

    cuda_error = cudaMemcpyToArray(cuArray,
                                      0,
                                      0,
                                      h_texture,
                                      LOOKUP_TABLE_LENGTH * sizeof(float),
                                      cudaMemcpyHostToDevice);
    if (cuda_error != cudaSuccess)
    {
        LOG("Failed copying texture array memory\n");
        return RET_ERROR_NO_MEM;
    }

    // Bind the array to the texture
    cuda_error = cudaBindTextureToArray(lookup_table, cuArray, channelDesc);
    if (cuda_error != cudaSuccess)
    {
        LOG("Failed binding texture array memory\n");
        return RET_ERROR_NO_MEM;
    }
    return RET_SUCCESS;
}



__global__
void scan_mode_4_kernel(frame_buffer_t * d_frame_buffers,
                        scan_output_buffer_t * d_output_buffer)
{
    const uint16_t * p_raw16_data = (uint16_t *)d_frame_buffers[0].d_image_buffer;
    const uint16_t * p_black_data = (uint16_t *)d_frame_buffers[1].d_image_buffer;
    const uint16_t * p_color_data = (uint16_t *)d_frame_buffers[2].d_image_buffer;

    const uint32_t NUM_SCAN_LINES_PER_IMAGE = d_frame_buffers[0].s_width / (sizeof(uint16_t) * 2);
    const uint32_t MAX_HEIGHT               = d_frame_buffers[0].s_height;
    const uint32_t SCATTER_DISTANCE         = d_output_buffer->p_config->img_processing_config.scatter_distance;
    const uint16_t PITCH                    = d_frame_buffers[0].s_width / sizeof(uint16_t);

    const uint32_t image_index = blockIdx.y;
    const uint32_t local_scan_line = blockIdx.x*blockDim.x + threadIdx.x;
    const uint32_t scan_line = image_index * NUM_SCAN_LINES_PER_IMAGE + local_scan_line;

    const uint32_t column = local_scan_line * 2;
    const uint32_t image_offset = image_index * PITCH * MAX_HEIGHT;
    const uint32_t scan_line_start_index = image_offset + column;

    if (local_scan_line >= NUM_SCAN_LINES_PER_IMAGE)
    {
        return;
    }

    float height = 0;
    uint32_t reflectance = 0;

    // First calculate the height based only on the red pixels
    uint32_t i, input_index;

    input_index = scan_line_start_index;

    for (i = 0; i < MAX_HEIGHT; i += 2) // only use data from each second line
    {
        int pixel_val = p_raw16_data[input_index] - p_black_data[input_index];
        if (pixel_val > 0)
        {
            float val = tex1D(lookup_table, pixel_val >> 4);

            reflectance += val;
            height      += i * val;
        }

        input_index += 2*PITCH; // jump 2 lines to the next line containing RED.
    }

    if (reflectance != 0)
    {
        height /= reflectance;
    }
    else
    {
        height = 0;
    }

    // normalize the height to its range
    d_output_buffer->p_height_buffer[scan_line] = (height/MAX_HEIGHT)*(1 << 16);
    d_output_buffer->p_reflectance_buffer[scan_line] = reflectance/(float)MAX_HEIGHT;

    if ( ((uint16_t)height & 1) == 0)
    {
        d_output_buffer->p_intensity_buffer[scan_line]
        = p_raw16_data[GET_INDEX(column, (uint16_t)height, image_offset)];
    }
    else
    {
        d_output_buffer->p_intensity_buffer[scan_line]
        = p_raw16_data[GET_INDEX(column, (uint16_t)height - 1, image_offset)] / 2
        + p_raw16_data[GET_INDEX(column, (uint16_t)height + 1, image_offset)] / 2;
    }


    const uint32_t scatter_height = MIN((uint16_t)height + SCATTER_DISTANCE, MAX_HEIGHT - 2);

    if ( (scatter_height & 1) == 0)
    {
        d_output_buffer->p_scatter_buffer[scan_line]
                = p_raw16_data[GET_INDEX(column, scatter_height, image_offset)];
    }
    else
    {
        d_output_buffer->p_scatter_buffer[scan_line]
              = p_raw16_data[GET_INDEX(column, scatter_height - 1, image_offset)] / 2
              + p_raw16_data[GET_INDEX(column, scatter_height + 1, image_offset)] / 2;
    }

    const uint32_t r_index = scan_line * 3;
    const uint32_t g_index = r_index + 1;
    const uint32_t b_index = r_index + 2;

    if (local_scan_line > 0 && local_scan_line < NUM_SCAN_LINES_PER_IMAGE - 1
                && (uint16_t)height > 0 && (uint16_t)height < MAX_HEIGHT - 1)
    {
        const uint32_t first_index = GET_INDEX(column - 1, (uint16_t)height - 1, image_offset);

        if ( ((uint16_t)height & 1) == 0)
        {
            d_output_buffer->p_rgb_buffer[r_index]
                  = p_color_data[first_index + PITCH + 1];

            d_output_buffer->p_rgb_buffer[g_index]
                  = p_color_data[first_index + PITCH] / 4
                  + p_color_data[first_index + PITCH + 2] / 4
                  + p_color_data[first_index + 1] / 4
                  + p_color_data[first_index + 2 * PITCH + 1] / 4;

            d_output_buffer->p_rgb_buffer[b_index]
                  = p_color_data[first_index] / 4
                  + p_color_data[first_index + 2] / 4
                  + p_color_data[first_index + 2 * PITCH] / 4
                  + p_color_data[first_index + 2 * PITCH + 2] / 4;
        }
        else
        {
            d_output_buffer->p_rgb_buffer[r_index]
                  = p_color_data[first_index + 1] / 2
                  + p_color_data[first_index + 2 * PITCH + 1] / 2;

            d_output_buffer->p_rgb_buffer[g_index]
                  = p_color_data[first_index + PITCH + 1];
            d_output_buffer->p_rgb_buffer[b_index]
                  = p_color_data[first_index + PITCH] / 2
                  + p_color_data[first_index + PITCH + 2] / 2;
        }
    }
}


