#include "camera.h"
#include "log.h"
#include "scanner.h"
#include "app_error.h"
#include "gpio.h"

#include "frame_buffer.h"
#include "img_processing.h"
#include "tcp_server.h"

#include <string.h>
#include <stdbool.h>
#include <byteswap.h>
#include <pthread.h>

#include <flycapture/C/FlyCapture2_C.h>
#include <flycapture/C/FlyCapture2Defs_C.h>

typedef enum
{
    IMAGE_CONTENT_LINE   = 0,
    IMAGE_CONTENT_BLANK  = 1,
    IMAGE_CONTENT_COLOR  = 2
} image_content_t;

typedef struct
{
    bool                     is_initialized;
    bool                     first_image_received;
    scan_mode_t              scan_mode;

    gpio_pin_t               scanner_ready_pin;
    float                    cycle_time_diff_threshold_ms;
    float                    cycle_time_ms;

    uint8_t                  img_type_index;
    uint8_t                  current_stream;

    image_info_t             image_info;

    img_processing_config_t  img_processing_config;

    stream_data_t           * p_stream_data;

    uint8_t *               * p_copy_of_last_images;
} scanner_t;


// The scan modes represent the length and type of the image
#define IMG_SEQ_LENGTH_OFFSET   0x80
#define IMG_SEQ_LENGTH_INDEX    0
#define IMG_SEQ_START_INDEX     1
#define IMG_SEQ_PADDING         0xFF

#define IMG_SEQ_0 {1 + IMG_SEQ_LENGTH_OFFSET, IMAGE_CONTENT_LINE,  IMG_SEQ_PADDING,     IMG_SEQ_PADDING}
#define IMG_SEQ_1 {2 + IMG_SEQ_LENGTH_OFFSET, IMAGE_CONTENT_LINE,  IMAGE_CONTENT_COLOR, IMG_SEQ_PADDING}
#define IMG_SEQ_2 {1 + IMG_SEQ_LENGTH_OFFSET, IMAGE_CONTENT_COLOR, IMG_SEQ_PADDING,     IMG_SEQ_PADDING}
#define IMG_SEQ_3 {2 + IMG_SEQ_LENGTH_OFFSET, IMAGE_CONTENT_LINE,  IMAGE_CONTENT_BLANK, IMG_SEQ_PADDING}
#define IMG_SEQ_4 {3 + IMG_SEQ_LENGTH_OFFSET, IMAGE_CONTENT_LINE,  IMAGE_CONTENT_BLANK, IMAGE_CONTENT_COLOR}

static const uint8_t IMG_SEQUENCES[][4] = {IMG_SEQ_0,
                                           IMG_SEQ_1,
                                           IMG_SEQ_2,
                                           IMG_SEQ_3,
                                           IMG_SEQ_4};

#define IMG_SEQ_GET_LENGTH(scan_mode)         (IMG_SEQUENCES[(scan_mode)][IMG_SEQ_LENGTH_INDEX] - IMG_SEQ_LENGTH_OFFSET)
#define IMG_SEQ_GET_IMG_TYPE(scan_mode, idx)  ((image_content_t)IMG_SEQUENCES[(scan_mode)][(idx)+1])

#define STOP_SCANNER_ON_ERROR(ret_code, message)        \
        if (ret_code != RET_SUCCESS)                    \
        {                                               \
            LOG("Scanner error: %s\n", message);        \
            scanner_error_handler(ret_code);            \
            return NULL;                                \
        }


static volatile bool m_continue_scan;
static scanner_t m_scanner;
static pthread_t m_scanner_thread;


// Declaration of auxiliary functions
static ret_code_t validate_image_format(fc2Image* p_img);
static ret_code_t get_diff_to_previous(fc2Image * pImage, float * p_diff_ms);
static ret_code_t get_image_content(uint8_t * p_img, image_content_t * p_image_content);
static ret_code_t validate_scan_config(scan_config_t * p_config);
static ret_code_t init_stream_data(scan_config_t * p_config);
static void       deinit_stream_data(void);
static ret_code_t init_previous_buffered_frames(uint8_t image_sequence_length, uint32_t image_size);
static void       deinit_previous_buffered_frames();


static void scanner_error_handler(ret_code_t ret_code)
{
    message_t message;

    message.type = MSG_TYPE_SCAN_ERROR;
    message.request.payload_len = 4;
    message.request.p_payload = (uint8_t *)&ret_code;

    scanner_stop();

    // The camera stop capture must be called.
    camera_stop_capture();

    tcp_server_send_message(&message);
}



static ret_code_t add_image_to_analysis(uint8_t * p_img_data, image_content_t img_type)
{
    ret_code_t ret_code;
    frame_buffer_t * const p_frame_buffer =
                    &m_scanner.p_stream_data[m_scanner.current_stream].p_frame_buffers[m_scanner.img_type_index];

    if (p_frame_buffer->num_images < p_frame_buffer->max_images - 1)
    {
        // The frame buffer is not yet full. Continue to buffer images.
        ret_code = frame_buffer_add_image(p_frame_buffer, p_img_data);
        if (ret_code != RET_SUCCESS)
        {
            LOG("Failed to add image\n");
            return RET_ERROR_NO_MEM;
        }
    }
    else
    {
        // The frame buffer is almost full. Add the last image
        ret_code = frame_buffer_add_image(p_frame_buffer, p_img_data);
        if (ret_code != RET_SUCCESS)
        {
            // This error cannot be recovered
            LOG_DEBUG("Failed to add last image\n");
            return RET_ERROR_NO_MEM;
        }

        // Reset the counter of the frame buffer.
        frame_buffer_clear(p_frame_buffer);

        // If the last image in the image sequence is received, the image processing can start.
        if (img_type == IMG_SEQ_GET_IMG_TYPE(m_scanner.scan_mode, IMG_SEQ_GET_LENGTH(m_scanner.scan_mode) - 1))
        {
            ret_code = img_processing_start(&m_scanner.p_stream_data[m_scanner.current_stream]);
            if (ret_code != RET_SUCCESS)
            {
                // This error cannot be recovered
                LOG("Failed to start image processing\n");
                return RET_ERROR_OTHER;
            }

            // Switch to next stream. Notice that this is only done for the last image in the sequence.
            m_scanner.current_stream = (m_scanner.current_stream + 1)
                    & (m_scanner.img_processing_config.num_streams - 1);

            // It is necessary to synchronize at this point to prevent the timing framework from being called when not finished
            cudaError_t cuda_error;
            cuda_error = cudaStreamSynchronize(m_scanner.p_stream_data[m_scanner.current_stream].stream);
            if (cuda_error != cudaSuccess)
            {
                LOG("cuda failed synchronizing: %s\n", cudaGetErrorString(cuda_error));
                return RET_ERROR_OTHER;
            }
        }
    }

    return RET_SUCCESS;
}


static ret_code_t process_image(fc2Image * p_new_image)
{
    ret_code_t ret_code;
    image_content_t image_content;
    float diff_ms;
    const image_content_t expected_image_type = IMG_SEQ_GET_IMG_TYPE(m_scanner.scan_mode,
                                                                     m_scanner.img_type_index);

    ret_code = validate_image_format(p_new_image);
    if (ret_code != RET_SUCCESS)
    {
        LOG("Invalid image format\n");
        return RET_ERROR_OTHER;
    }

    ret_code = get_diff_to_previous(p_new_image, &diff_ms);
    if (ret_code != RET_SUCCESS)
    {
        LOG("Failed to get image capture time diff\n");
        return RET_ERROR_OTHER;
    }

    if (diff_ms < 0)
    {
        LOG("The returned diff between the images is negative\n");
        return RET_ERROR_OTHER;
    }

    if (diff_ms > m_scanner.cycle_time_ms * 50)
    {
        LOG("More than 50 frames dropped. Unlikely to recover. Diff is %f\n", diff_ms);
        return RET_ERROR_OTHER;
    }

    if (abs(diff_ms - m_scanner.cycle_time_ms) < m_scanner.cycle_time_diff_threshold_ms)
    {
        // No frames are dropped. Check image content and store a backup of the image.
        ret_code = get_image_content(p_new_image->pData, &image_content);
        if (ret_code != RET_SUCCESS)
        {
            LOG("Failed to retrieve image content\n");
            return RET_ERROR_OTHER;
        }


        if (image_content == expected_image_type)
        {
#if defined(SCANNER_BUFFER_ASYNC)
            cudaError_t cuda_error;
            cuda_error = cudaMemcpyAsync(m_scanner.p_copy_of_last_images[m_scanner.img_type_index],
                                         p_new_image->pData,
                                         m_scanner.image_info.frame_height * m_scanner.image_info.num_scan_lines_per_image * 2 * sizeof(uint16_t),
                                         cudaMemcpyHostToHost,
                                         m_scanner.p_stream_data[m_scanner.current_stream].stream);

            if (cuda_error != cudaSuccess)
            {
                LOG("Failed storing backup image\n");
                return RET_ERROR_OTHER;
            }
#else
            memcpy(m_scanner.p_copy_of_last_images[m_scanner.img_type_index],
                   p_new_image->pData,
                   m_scanner.image_info.frame_height * m_scanner.image_info.num_scan_lines_per_image * 2 * sizeof(uint16_t));
#endif

        }
        else
        {
            LOG("Unexpected image content arrived. Got %d, expected %d\n", image_content, expected_image_type);
            return RET_ERROR_OTHER;
        }
    }
    else
    {
        LOG_DEBUG("Frames dropped. Previously buffered frames will be used. Diff is %f\n", diff_ms);
    }

    // Continue to add images to analysis until all expected images are handled.
    while ( diff_ms > m_scanner.cycle_time_diff_threshold_ms )
    {
        // It is valid to always use the backup images
        ret_code = add_image_to_analysis(m_scanner.p_copy_of_last_images[m_scanner.img_type_index], expected_image_type);
        if (ret_code != RET_SUCCESS)
        {
            LOG("Failed to add image to analysis\n");
            return RET_ERROR_OTHER;
        }

        // switch to the next image type in the image sequence
        if (m_scanner.img_type_index < IMG_SEQ_GET_LENGTH(m_scanner.scan_mode) - 1)
        {
            m_scanner.img_type_index++;
        }
        else
        {
            m_scanner.img_type_index = 0;
        }

        // subtract the cycle time to see if there are more images that should be processed
        diff_ms -= m_scanner.cycle_time_ms;
    }

    return RET_SUCCESS;
}



static void * scan_loop(void * args)
{
    ret_code_t ret_code;
    fc2Image * p_image;

    ret_code = camera_start_capture();
    STOP_SCANNER_ON_ERROR(ret_code, "Start capture");

    gpio_deinit_output(m_scanner.scanner_ready_pin);

    ret_code = gpio_init_output(m_scanner.scanner_ready_pin);
    STOP_SCANNER_ON_ERROR(ret_code, "GPIO init");

    ret_code = gpio_pin_set(m_scanner.scanner_ready_pin, 1);
    STOP_SCANNER_ON_ERROR(ret_code, "Pin set, high, first time");

    LOG_DEBUG("Starting to retrieve images\n");

    bool sequence_started = false;
    image_content_t image_content, first_image_type = IMG_SEQ_GET_IMG_TYPE(m_scanner.scan_mode, 0);

    while (m_continue_scan)
    {
        ret_code = camera_get_buffer(&p_image);
        STOP_SCANNER_ON_ERROR(ret_code, "Retrieve camera buffer");

        ret_code = gpio_pin_set(m_scanner.scanner_ready_pin, 0);
        STOP_SCANNER_ON_ERROR(ret_code, "GPIO set 0");

        if (sequence_started == false)
        {
            ret_code = get_image_content(p_image->pData, &image_content);
            STOP_SCANNER_ON_ERROR(ret_code, "Retrieve image content while synching");

            if (image_content == first_image_type)
            {
                LOG_DEBUG("Sequence synchronization finished\n");
                sequence_started = true;
            }
        }

        if (sequence_started == true)
        {
            ret_code = process_image(p_image);
            STOP_SCANNER_ON_ERROR(ret_code, "Processing");
        }

        ret_code = gpio_pin_set(m_scanner.scanner_ready_pin, 1);
        STOP_SCANNER_ON_ERROR(ret_code, "GPIO set 1");
    }

    ret_code = gpio_pin_set(m_scanner.scanner_ready_pin, 0);
    STOP_SCANNER_ON_ERROR(ret_code, "GPIO set 0, last time");

    ret_code = camera_stop_capture();
    STOP_SCANNER_ON_ERROR(ret_code, "Stop capture");

    return NULL;
}


ret_code_t scanner_start(void)
{
    if (!m_scanner.is_initialized)
    {
        return RET_ERROR_NOT_INITIALIZED;
    }

    m_scanner.current_stream = 0;
    m_scanner.img_type_index = 0;
    m_scanner.first_image_received = false;

    // reset frame buffers
    for (int i = 0; i < m_scanner.img_processing_config.num_streams; i++)
    {
        for (int j = 0; j < IMG_SEQ_GET_LENGTH(m_scanner.scan_mode); j++)
        {
            frame_buffer_clear(&m_scanner.p_stream_data[i].p_frame_buffers[j]);
        }
    }

    m_continue_scan = true;
    int err;
    err = pthread_create(&m_scanner_thread, NULL, scan_loop, NULL);
    if (err != 0)
    {
        perror("Creating thread\n");
        return RET_ERROR_OTHER;
    }

    return RET_SUCCESS;
}



ret_code_t scanner_stop(void)
{
    if (!m_scanner.is_initialized)
    {
        return RET_ERROR_NOT_INITIALIZED;
    }

    m_continue_scan = false;

    pthread_join(m_scanner_thread, NULL);

    return RET_SUCCESS;
}

ret_code_t scanner_set_config(scan_config_t * p_config)
{
    if (p_config == NULL)
    {
        return RET_ERROR_NULL;
    }

    ret_code_t ret_code;

    ret_code = validate_scan_config(p_config);    gpio_deinit_output(m_scanner.scanner_ready_pin);
    gpio_init_output(m_scanner.scanner_ready_pin);
    if (ret_code != RET_SUCCESS)
    {
        LOG("Invalid config\n");
        return RET_ERROR_INVALID_CONFIG;
    }

    ret_code = init_stream_data(p_config);
    RETURN_ON_ERROR(ret_code);

    ret_code = init_previous_buffered_frames(IMG_SEQ_GET_LENGTH(p_config->scan_mode),
                                             p_config->image_info.frame_height
                                             * p_config->image_info.num_scan_lines_per_image * 2 * sizeof(uint16_t));
    RETURN_ON_ERROR(ret_code);

    ret_code = img_processing_set_config(p_config);
    RETURN_ON_ERROR(ret_code);



    m_scanner.scanner_ready_pin = p_config->scanner_ready_pin;
    m_scanner.scan_mode = p_config->scan_mode;
    m_scanner.img_type_index = 0;
    m_scanner.cycle_time_diff_threshold_ms = p_config->cycle_time_diff_threshold_ms;
    m_scanner.cycle_time_ms = 1000.0 / p_config->fps;

    memcpy(&m_scanner.image_info, &p_config->image_info, sizeof(image_info_t));
    memcpy(&m_scanner.img_processing_config, &p_config->img_processing_config, sizeof(img_processing_config_t));

    m_scanner.is_initialized = true;

    LOG_DEBUG("Scan mode is now set to %d\n", m_scanner.scan_mode);

    return RET_SUCCESS;
}

static ret_code_t validate_image_format(fc2Image* p_img)
{
    if (p_img->format != FC2_PIXEL_FORMAT_RAW16)
    {
        LOG_DEBUG("Unexpected pixel format\n");
        return RET_ERROR_INVALID_CONFIG;
    }

    if (p_img->bayerFormat != FC2_BT_RGGB)
    {
        LOG_DEBUG("Unexpected bayer tile format\n");
        return RET_ERROR_INVALID_CONFIG;
    }

    if (p_img->rows != m_scanner.image_info.frame_height)
    {
        LOG_DEBUG("Unexpected height. Got %d, expected %d\n",p_img->rows, m_scanner.image_info.frame_height);
        return RET_ERROR_INVALID_CONFIG;
    }

    // Because of the bayer format, there should be twice as many cols as scan lines.
    if (p_img->cols != m_scanner.image_info.num_scan_lines_per_image * 2)
    {
        LOG_DEBUG("Unexpected width. Got %d, expected %d\n", p_img->cols, m_scanner.image_info.num_scan_lines_per_image * 2);
        return RET_ERROR_INVALID_CONFIG;
    }

    // Because of the RAW16 format, the stride should be twice as many cols.
    if (p_img->stride != m_scanner.image_info.num_scan_lines_per_image * 4)
    {
        LOG_DEBUG("Unexpected stride. Got %d, expected %d\n", p_img->stride, m_scanner.image_info.num_scan_lines_per_image * 4);
        return RET_ERROR_INVALID_CONFIG;
    }

    return RET_SUCCESS;
}


static ret_code_t get_diff_to_previous(fc2Image * pImage, float * p_diff_ms)
{
    static fc2TimeStamp prev_time_stamp;

    if (m_scanner.first_image_received == false)
    {
        // Reset prev_time_stamp
        prev_time_stamp = fc2GetImageTimeStamp(pImage);

        m_scanner.first_image_received = true;

        *p_diff_ms = m_scanner.cycle_time_ms;
        return RET_SUCCESS;
    }

    fc2TimeStamp      curr_time_stamp = fc2GetImageTimeStamp(pImage);

    int secondsDiff = curr_time_stamp.cycleSeconds - prev_time_stamp.cycleSeconds;
    int countDiff   = curr_time_stamp.cycleCount - prev_time_stamp.cycleCount;

    if (secondsDiff < 0)
    {
        secondsDiff = 128 + secondsDiff;
    }

    if (countDiff < 0)
    {
        // There are 8000 cycles in a second
        countDiff = 8000 + countDiff;

        secondsDiff--;

    }

    *p_diff_ms = double(secondsDiff)*1000 + double(countDiff) / 8;

    memcpy(&prev_time_stamp, &curr_time_stamp, sizeof(fc2TimeStamp));

    return RET_SUCCESS;
}


static ret_code_t get_image_content(uint8_t * p_img, image_content_t * p_image_content)
{
    uint32_t gpio_state = *(uint32_t*)(p_img + 4);

    *p_image_content = (image_content_t)((gpio_state >> 4) & 3);

    return RET_SUCCESS;
}



static ret_code_t validate_scan_config(scan_config_t * p_config)
{
    if (   p_config->image_info.frame_height             == 0
        || p_config->image_info.num_scan_lines_per_image == 0)
    {
        return RET_ERROR_INVALID_CONFIG;
    }

    if (p_config->image_info.num_scan_lines_per_image > 1920 / 2)
    {
        LOG_DEBUG("Too many scan lines per image\n");
        return RET_ERROR_INVALID_CONFIG;
    }

    if (p_config->cycle_time_diff_threshold_ms == 0)
    {
        LOG("Cycle time diff threshold cannot be 0\n");
        return RET_ERROR_INVALID_CONFIG;
    }

    if (p_config->fps == 0)
    {
        LOG("Fps cannot be set to 0\n");
        return RET_ERROR_INVALID_CONFIG;
    }

    return RET_SUCCESS;
}


static ret_code_t init_stream_data(scan_config_t * p_config)
{
    deinit_stream_data();

    m_scanner.p_stream_data = (stream_data_t*)malloc(p_config->img_processing_config.num_streams * sizeof(stream_data_t));
    if (m_scanner.p_stream_data == NULL)
    {
        LOG_DEBUG("Failed to allocate data for streams\n");
        return RET_ERROR_NO_MEM;
    }

    uint8_t i;
    ret_code_t ret_code;
    for (i = 0; i < p_config->img_processing_config.num_streams; i++)
    {
        memset(&m_scanner.p_stream_data[i], 0, sizeof(stream_data_t));
        ret_code = stream_data_init(&m_scanner.p_stream_data[i],
                                    IMG_SEQ_GET_LENGTH(p_config->scan_mode),
                                    p_config);
        RETURN_ON_ERROR(ret_code);
    }

    return RET_SUCCESS;
}

static void deinit_stream_data(void)
{
    if (m_scanner.p_stream_data != NULL)
    {
        uint8_t i;
        for (i = 0; i < m_scanner.img_processing_config.num_streams; i++)
        {
            stream_data_deinit(&m_scanner.p_stream_data[i], IMG_SEQ_GET_LENGTH(m_scanner.scan_mode));
        }
        free(m_scanner.p_stream_data);
    }
}

static ret_code_t init_previous_buffered_frames(uint8_t image_sequence_length, uint32_t image_size)
{
    deinit_previous_buffered_frames();

    m_scanner.p_copy_of_last_images = (uint8_t **) malloc(image_sequence_length * sizeof(uint8_t*));
    if (m_scanner.p_copy_of_last_images == NULL)
    {
        return RET_ERROR_NO_MEM;
    }

    uint8_t i;
    for (i = 0; i < image_sequence_length; i++)
    {
        cudaError_t cuda_error = cudaMallocHost(&m_scanner.p_copy_of_last_images[i], image_size);
        if (cuda_error != cudaSuccess)
        {
            return RET_ERROR_NO_MEM;
        }
    }

    return RET_SUCCESS;
}

static void deinit_previous_buffered_frames()
{
    if (m_scanner.p_copy_of_last_images != NULL)
    {
        uint8_t i;
        for (i = 0; i < IMG_SEQ_GET_LENGTH(m_scanner.scan_mode); i++)
        {
            if (m_scanner.p_copy_of_last_images[i] != NULL)
            {
                cudaFree(m_scanner.p_copy_of_last_images[i]);
            }
        }

        free(m_scanner.p_copy_of_last_images);
    }
}

