#ifndef SCANNER_H__
#define SCANNER_H__

#ifdef __cplusplus
extern "C"
{
#endif

#include <stdint.h>
#include "ret_code.h"
#include "gpio.h"

#if !defined(SCANNER_BUFFER_ASYNC) && \
    !defined(SCANNER_BUFFER_SYNC)
#error "Specify scanner buffer configuration"
#endif

/**@brief The defined scan modes.
 *
 * @details Up to 2^32 scan modes can be defined. In the description the frame types are abbreviated as following:
 *
 * L: An image containing the laser line.
 * B: An image where the laser line and LED is turned off.
 * C: An illuminated by LEDs.
 */
typedef enum
{
    SCAN_MODE_0 = 0,    /**< L, L, L, ... */
    SCAN_MODE_1 = 1,    /**< L, C, L, C, ... */
    SCAN_MODE_2 = 2,    /**< C, C, C, ... */
    SCAN_MODE_3 = 3,    /**< L, B, L, B, ... */
    SCAN_MODE_4 = 4,    /**< L, B, C, L, B, C, ... */

    SCAN_MODE_FORCE_32_BIT = 0xFFFFFFFF
} scan_mode_t;

/**@brief The image info structure holds information about image configuration. */
typedef struct image_info_s
{
    uint32_t            frame_height;               /**< The height of the image in pixels. */
    uint32_t            num_scan_lines_per_image;   /**< The number of scan lines per image. That is the number of columns divided by two. */
} image_info_t;


/** @brief Image processing configuration. */
typedef struct
{
    uint32_t scatter_distance;      /**< Scatter distance is the intensity of the pixel calculated some distance from the height. Defined in pixels. */
    uint32_t threshold_value;       /**< The threshold value sets a lower limit of pixel intensities. All values below are not used for height and reflectance calculations. */
    float    pixel_power;           /**< The exponent used for height calculation. */

    uint32_t num_streams;           /**< The number of CUDA streams used for image processing. */
    uint32_t threads_per_block;     /**< The threads per block is a kernel launch parameter. */
    uint32_t frames_per_message;    /**< Frames per message indicates the size of the frame buffer and output message. */
} img_processing_config_t;


/** @brief Scanner configuration. */
typedef struct scan_config_s
{
    scan_mode_t              scan_mode;                     /**< The active scan mode used to retrieve the image parameters. */
    image_info_t             image_info;                    /**< The image size used for parameter extraction. */
    gpio_pin_t               scanner_ready_pin;             /**< The GPIO indicating when the scanner framework is busy. */
    float                    fps;                           /**< The expected camera frame rate. This is used to see if frames are dropped. */
    float                    cycle_time_diff_threshold_ms;  /**< A threshold value determining the maximum jitter in image reception time.*/

    img_processing_config_t  img_processing_config;         /**< The image processing configuration used for parameter extraction. */
} scan_config_t;


/** @brief A structure holding a pointer to the output buffer.
 *
 *  @details    The structure does also hold pointers to sub-buffers of the output buffer.
 *              These pointers are only calculated once, reducing the computational load.
 * */
typedef struct scan_output_buffer_s
{
    uint8_t                 * p_buffer;                 /**< Pointer to the raw output buffer. */
    uint32_t                buffer_size;                /**< The size of the output buffer in bytes. */

    scan_config_t           * p_config;                 /**< Pointer to the configuration field of the output buffer. */

    uint16_t               * p_height_buffer;           /**< Height is stored in a short word. The field is num_scan_lines_per_image*frames_per_message long. */
    uint16_t               * p_intensity_buffer;        /**< Pointer to the location where intensity data is stored. It represents the intensity of the laser line. */
    uint16_t               * p_scatter_buffer;          /**< The scatter is the intensity some distance from the laser line. */
    uint16_t               * p_reflectance_buffer;      /**< The reflectance is the average intensity of a scan column. */
    uint16_t               * p_rgb_buffer;              /**< The RGB buffer stores the RGB values in a 16-bit triplet. */
    float                  * p_gpu_execution_time;      /**< The previous GPU execution time measured in milliseconds. */
    float                  * p_total_execution_time;    /**< The previous GPU + TCP execution time measured in milliseconds. */
} scan_output_buffer_t;


/** @brief Starts scanning.
 *
 *  @details Starts a thread polling the camera for images. When images are received, they are validated and buffered.
 *           When the buffer is full, an image processing task is started.
 *
 *  @return RET_SUCCESS                 If scanning was started successfully.
 *  @return RET_ERROR_NOT_INITIALIZED   If the scanner or image processing is not initialized.
 *  @return RET_ERROR_OTHER             If the start thread procedure failed.
 */
ret_code_t scanner_start(void);


/** @brief Stops scanning.
 *
 *  @return RET_SUCCESS                 If scanning was started successfully.
 *  @return RET_ERROR_NOT_INITIALIZED   If the scanner or image processing is not initialized.
 */
ret_code_t scanner_stop(void);


/** @brief Validates and sets the configuration.
 *
 *  @param[in]                          Pointer to scanner configuration.
 *
 *  @return RET_SUCCESS                 If scanning was started successfully.
 *  @return RET_ERROR_NO_MEM            If memory management failed.
 *  @return RET_ERROR_INVALID_PARAMETER If the current configuration was invalid.
 */
ret_code_t scanner_set_config(scan_config_t * p_config);

#ifdef __cplusplus
}
#endif



#endif // SCANNER_H__
