#include "state_machine.h"

#include <string.h>

#include "app_error.h"
#include "scanner.h"
#include "camera.h"
#include "config_evt_broker.h"
#include <flycapture/C/FlyCapture2_C.h>
#include <flycapture/C/FlyCapture2Defs_C.h>

static state_t m_current_state;

ret_code_t state_machine_init(void)
{
    m_current_state = STATE_IDLE;
    return RET_SUCCESS;
}

static ret_code_t on_goto_state(tcp_evt_t const * p_evt, message_t * p_response)
{
    ret_code_t ret_code = RET_SUCCESS;
    uint32_t   new_state;


    if (p_evt->message.request.payload_len != 4)
    {
        return RET_ERROR_INVALID_LENGTH;
    }

    switch (m_current_state)
    {
        case STATE_IDLE:
        case STATE_CONFIG:
            // no implementation
            break;
        case STATE_SCANNING:
            ret_code = scanner_stop();
    }

    RETURN_ON_ERROR(ret_code);
    m_current_state = STATE_IDLE;

    // copy in new state
    memcpy(&new_state, p_evt->message.request.p_payload, 4);

    switch ((state_t)new_state)
    {
        case STATE_IDLE:
        case STATE_CONFIG:
            // no implementation
            break;
        case STATE_SCANNING:
            ret_code = scanner_start();
            break;
        default:
            return RET_ERROR_INVALID_STATE;
    }

    RETURN_ON_ERROR(ret_code);
    m_current_state = (state_t)new_state;
    LOG_DEBUG("State changed to %d\n", m_current_state);

    return RET_SUCCESS;
}

static ret_code_t on_config_evt(tcp_evt_t const * p_evt, message_t * p_response)
{
    if (m_current_state != STATE_CONFIG)
    {
        return RET_ERROR_INVALID_STATE;
    }

    return config_evt_broker_on_tcp_evt(p_evt, p_response);
}

static void send_image_get_error_message(ret_code_t ret_code)
{
    message_t msg = {0};

    msg.type = MSG_TYPE_RESPONSE;
    msg.response.request_type = MSG_TYPE_GET_IMAGE;
    msg.response.ret_code = RET_ERROR_OTHER;

    tcp_server_send_message(&msg);
}


static void send_image(const char * name)
{
    FILE* fp;
    uint8_t buffer[1024];
    uint32_t size;

    fp = fopen(name, "r");
    if (fp == NULL)
    {
        send_image_get_error_message(RET_ERROR_NO_MEM);
        return;
    }

    // first send header
    buffer[0] = MSG_TYPE_RESPONSE;

    fseek(fp, 0, SEEK_END); // seek to end of file
    size = ftell(fp); // get current file pointer
    fseek(fp, 0, SEEK_SET); // seek back to beginning of file

    size += 2; // add the size of the return code and request type

    memcpy(&buffer[1], &size, sizeof(size));
    buffer[5] = MSG_TYPE_GET_IMAGE;
    buffer[6] = RET_SUCCESS;

    tcp_server_send_data(buffer, 7);

    // then send the file
    size_t read;
    while ( (read = fread(&buffer, 1, sizeof(buffer), fp)) != 0)
    {
        tcp_server_send_data(buffer, read);
    }

    fclose(fp);
}




static void on_get_image(void)
{
    fc2Image    img;
    fc2Error    err_code;
    ret_code_t  ret_code;

    ret_code = camera_get_image(&img);
    if (ret_code != RET_SUCCESS)
    {
        send_image_get_error_message(ret_code);
        return;
    }

    err_code = fc2SaveImage(&img, "my_img.png", FC2_PNG);
    if (err_code != FC2_ERROR_OK)
    {
        send_image_get_error_message(RET_ERROR_NO_MEM);
        return;
    }

    send_image("my_img.png");

    camera_destroy_image(&img);

}



void state_machine_on_tcp_evt(tcp_evt_t const * p_evt)
{
    ret_code_t  ret_code = RET_SUCCESS;
    message_t   response = {0};

    switch (p_evt->type)
    {

        case TCP_EVT_TYPE_CONNECTED:
        case TCP_EVT_TYPE_DISCONNECTED:
            switch (m_current_state)
            {
                case STATE_IDLE:
                case STATE_CONFIG:
                    // no implementation
                    break;
                case STATE_SCANNING:
                    ret_code = scanner_stop();
                    break;
            }

           APP_ERROR_CHECK(ret_code);

           m_current_state = STATE_IDLE;
           LOG_DEBUG("Connection state changed, returned to IDLE\n");
           break;

        case TCP_EVT_TYPE_RECV_MESSAGE:
            response.type = MSG_TYPE_RESPONSE;
            response.response.request_type = p_evt->message.type;

            switch (p_evt->message.type)
            {
                case MSG_TYPE_GOTO_STATE:
                    ret_code = on_goto_state(p_evt, &response);
                    response.response.ret_code = ret_code;
                    tcp_server_send_message(&response);
                    break;

                case MSG_TYPE_CONFIG_CAMERA:
                case MSG_TYPE_CONFIG_SCAN:
                    ret_code = on_config_evt(p_evt, &response);
                    response.response.ret_code = ret_code;
                    tcp_server_send_message(&response);
                    break;

                case MSG_TYPE_GET_IMAGE:
                    on_get_image();
                    break;

                default:
                    ret_code = RET_ERROR_NOT_IMPLEMENTED;
                    break;
            }


    }
}
