#include "stream_data.h"
#include "app_error.h"


#define DEINIT_AND_RETURN_ON_FAILURE(ret_code)  \
    if (ret_code != RET_SUCCESS)                \
    {                                           \
        stream_data_deinit(p_stream_data, seq_length);      \
        return ret_code;                        \
    }

static ret_code_t init_stream(cudaStream_t * p_stream);
static void       deinit_stream(cudaStream_t * p_stream);


static ret_code_t init_frame_buffers(frame_buffer_t ** pp_frame_buffers,
                                     frame_buffer_t ** pd_frame_buffers,
                                     uint8_t seq_length,
                                     cudaStream_t * p_stream,
                                     image_info_t * p_image_info,
                                     uint8_t size);
static void       deinit_frame_buffers(frame_buffer_t ** pp_frame_buffers,
                                       frame_buffer_t ** pd_frame_buffers,
                                       uint8_t seq_length);

#ifdef STREAM_DATA_OUTPUT_BUFFER_HOST
static ret_code_t init_output_buffers_host(scan_output_buffer_t ** pph_buffer,
                                           scan_output_buffer_t ** ppd_buffer,
                                           uint8_t ** ppd_buffer_raw,
                                           scan_config_t * p_config);
static void       deinit_output_buffers_host(scan_output_buffer_t ** pph_buffer,
                                             scan_output_buffer_t ** ppd_buffer,
                                             uint8_t ** ppd_buffer_raw);
#endif

#ifdef STREAM_DATA_OUTPUT_BUFFER_DEVICE
static ret_code_t init_output_buffers_device(scan_output_buffer_t ** pph_buffer,
                                             scan_output_buffer_t ** ppd_buffer,
                                             uint8_t ** ppd_buffer_raw,
                                             scan_config_t * p_config);
static void       deinit_output_buffers_device(scan_output_buffer_t ** pph_buffer,
                                               scan_output_buffer_t ** ppd_buffer,
                                               uint8_t ** ppd_buffer_raw);
#endif


static ret_code_t init_event(cudaEvent_t * p_event);
static void deinit_event(cudaEvent_t * p_event);


ret_code_t stream_data_init(stream_data_t * p_stream_data,
                            uint8_t seq_length,
                            scan_config_t * p_config)
{
    ret_code_t ret_code;

    ret_code = init_stream(&p_stream_data->stream);
    DEINIT_AND_RETURN_ON_FAILURE(ret_code)

    ret_code = init_frame_buffers(&p_stream_data->p_frame_buffers,
                                  &p_stream_data->d_frame_buffers,
                                  seq_length,
                                  &p_stream_data->stream,
                                  &p_config->image_info,
                                  p_config->img_processing_config.frames_per_message);
    DEINIT_AND_RETURN_ON_FAILURE(ret_code)

#ifdef STREAM_DATA_OUTPUT_BUFFER_DEVICE
    ret_code = init_output_buffers_device(&p_stream_data->h_scan_output_buffer,
                                          &p_stream_data->d_scan_output_buffer,
                                          &p_stream_data->d_scan_output_buffer_raw,
                                          p_config);
#else
    ret_code = init_output_buffers_host(&p_stream_data->h_scan_output_buffer,
                                        &p_stream_data->d_scan_output_buffer,
                                        &p_stream_data->d_scan_output_buffer_raw,
                                        p_config);
#endif
    DEINIT_AND_RETURN_ON_FAILURE(ret_code)

    ret_code = init_event(&p_stream_data->start_event);
    DEINIT_AND_RETURN_ON_FAILURE(ret_code)

    ret_code = init_event(&p_stream_data->gpu_finished_event);
    DEINIT_AND_RETURN_ON_FAILURE(ret_code)

    ret_code = init_event(&p_stream_data->sending_finished_event);
    DEINIT_AND_RETURN_ON_FAILURE(ret_code)

    return RET_SUCCESS;
}


void stream_data_deinit(stream_data_t * p_stream_data, uint8_t seq_length)
{
    deinit_stream(&p_stream_data->stream);
    deinit_frame_buffers(&p_stream_data->p_frame_buffers,
                         &p_stream_data->d_frame_buffers,
                         seq_length);
#ifdef STREAM_DATA_OUTPUT_BUFFER_DEVICE
    deinit_output_buffers_device(&p_stream_data->h_scan_output_buffer,
                                 &p_stream_data->d_scan_output_buffer,
                                 &p_stream_data->d_scan_output_buffer_raw);
#else
    deinit_output_buffers_host(&p_stream_data->h_scan_output_buffer,
                               &p_stream_data->d_scan_output_buffer,
                               &p_stream_data->d_scan_output_buffer_raw);
#endif
    deinit_event(&p_stream_data->start_event);
    deinit_event(&p_stream_data->gpu_finished_event);
    deinit_event(&p_stream_data->sending_finished_event);
}

static ret_code_t init_stream(cudaStream_t * p_stream)
{
    cudaError_t cuda_error;
    cuda_error = cudaStreamCreateWithFlags(p_stream, cudaStreamNonBlocking);
    if (cuda_error != cudaSuccess)
    {
        return RET_ERROR_NO_MEM;
    }
    return RET_SUCCESS;
}

static void deinit_stream(cudaStream_t * p_stream)
{
    if (*p_stream != 0)
    {
        cudaStreamDestroy(*p_stream);
        *p_stream = 0;
    }
}

static ret_code_t init_frame_buffers(frame_buffer_t ** pp_frame_buffers,
                                     frame_buffer_t ** pd_frame_buffers,
                                     uint8_t seq_length,
                                     cudaStream_t * p_stream,
                                     image_info_t * p_image_info,
                                     uint8_t size)
{
    ret_code_t ret_code;
    cudaError_t cuda_error;

    *pp_frame_buffers = (frame_buffer_t *)malloc(seq_length * sizeof(frame_buffer_t));
    if (*pp_frame_buffers == NULL)
    {
        return RET_ERROR_NO_MEM;
    }

    cuda_error = cudaMalloc(pd_frame_buffers, seq_length * sizeof(frame_buffer_t));
    if (cuda_error != cudaSuccess)
    {
        *pd_frame_buffers = NULL;
        return RET_ERROR_NO_MEM;
    }

    frame_buffer_init_t init;
    init.max_images   = size;
    init.image_height = p_image_info->frame_height;
    init.p_stream     = p_stream;

    // The number of scan lines is only half of the columns
    init.image_width  = p_image_info->num_scan_lines_per_image * sizeof(uint16_t) * 2;

    uint8_t i;
    for (i = 0; i < seq_length; i++)
    {
        ret_code = frame_buffer_init(&(*pp_frame_buffers)[i], &init);
        RETURN_ON_ERROR(ret_code);

        // copy the state information
        cuda_error = cudaMemcpy(&(*pd_frame_buffers)[i],
                                &(*pp_frame_buffers)[i],
                                sizeof(frame_buffer_t),
                                cudaMemcpyHostToDevice);
        if (cuda_error != cudaSuccess)
        {
            return RET_ERROR_NO_MEM;
        }
    }

    return RET_SUCCESS;
}

static void deinit_frame_buffers(frame_buffer_t ** pp_frame_buffers,
                                 frame_buffer_t ** pd_frame_buffers,
                                 uint8_t seq_length)
{
    if (*pp_frame_buffers != NULL)
    {
        uint8_t i;
        for (i = 0; i < seq_length; i++)
        {
            frame_buffer_deinit(&((*pp_frame_buffers)[i]));
        }
    }

    if (*pd_frame_buffers != NULL)
    {
        cudaFree(*pd_frame_buffers);
        *pd_frame_buffers = NULL;
    }

    *pp_frame_buffers = NULL;
}


#ifdef STREAM_DATA_OUTPUT_BUFFER_DEVICE

static ret_code_t init_output_buffers_device(scan_output_buffer_t ** pph_buffer,
                                             scan_output_buffer_t ** ppd_buffer,
                                             uint8_t ** ppd_buffer_raw,
                                             scan_config_t * p_config)
{
    cudaError_t cuda_error;

    cuda_error = cudaMallocHost(pph_buffer, sizeof(scan_output_buffer_t));
    if (cuda_error != cudaSuccess)
    {
        *pph_buffer = NULL;
        return RET_ERROR_NO_MEM;
    }

    cuda_error = cudaMalloc(ppd_buffer, sizeof(scan_output_buffer_t));
    if (cuda_error != cudaSuccess)
    {
        *ppd_buffer = NULL;
        return RET_ERROR_NO_MEM;
    }

    const uint32_t SCAN_LINES_PER_IMAGE = p_config->image_info.num_scan_lines_per_image;
    const uint32_t SCAN_LINES           = p_config->img_processing_config.frames_per_message * SCAN_LINES_PER_IMAGE;

    // first calculate size.
    (*pph_buffer)->buffer_size = 0;

    (*pph_buffer)->buffer_size += sizeof(scan_config_t);
    (*pph_buffer)->buffer_size += sizeof(*(*pph_buffer)->p_height_buffer) * SCAN_LINES;
    (*pph_buffer)->buffer_size += sizeof(*(*pph_buffer)->p_intensity_buffer) * SCAN_LINES;
    (*pph_buffer)->buffer_size += sizeof(*(*pph_buffer)->p_reflectance_buffer) * SCAN_LINES;
    (*pph_buffer)->buffer_size += sizeof(*(*pph_buffer)->p_scatter_buffer) * SCAN_LINES;
    (*pph_buffer)->buffer_size += sizeof(*(*pph_buffer)->p_rgb_buffer) * SCAN_LINES * 3;
    (*pph_buffer)->buffer_size += sizeof(*(*pph_buffer)->p_gpu_execution_time);
    (*pph_buffer)->buffer_size += sizeof(*(*pph_buffer)->p_total_execution_time);

    // then allocate
    cuda_error = cudaMallocHost(&(*pph_buffer)->p_buffer,(*pph_buffer)->buffer_size);
    if (cuda_error != cudaSuccess)
    {
        (*pph_buffer)->p_buffer = NULL;
        return RET_ERROR_NO_MEM;
    }

    cuda_error = cudaMalloc(ppd_buffer_raw,(*pph_buffer)->buffer_size);
    if (cuda_error != cudaSuccess)
    {
        ppd_buffer_raw = NULL;
        return RET_ERROR_NO_MEM;
    }

    // Set the corresponding pointers of the buffers.
    scan_output_buffer_t temp_buffer; // Temporary structure to point to the device buffer
    temp_buffer.p_buffer    = *ppd_buffer_raw;
    temp_buffer.buffer_size = (*pph_buffer)->buffer_size;

    uint32_t index = 0;
    (*pph_buffer)->p_config = (scan_config_t *)&(*pph_buffer)->p_buffer[index];
    temp_buffer.p_config    = (scan_config_t *)&temp_buffer.p_buffer[index];
    index += sizeof(scan_config_t);

    (*pph_buffer)->p_height_buffer = (uint16_t *)&(*pph_buffer)->p_buffer[index];
    temp_buffer.p_height_buffer    = (uint16_t *)&temp_buffer.p_buffer[index];
    index += sizeof(*(*pph_buffer)->p_height_buffer) * SCAN_LINES;

    (*pph_buffer)->p_intensity_buffer = (uint16_t *)&(*pph_buffer)->p_buffer[index];
    temp_buffer.p_intensity_buffer    = (uint16_t *)&temp_buffer.p_buffer[index];
    index += sizeof(*(*pph_buffer)->p_intensity_buffer) * SCAN_LINES;

    (*pph_buffer)->p_scatter_buffer = (uint16_t *)&(*pph_buffer)->p_buffer[index];
    temp_buffer.p_scatter_buffer    = (uint16_t *)&temp_buffer.p_buffer[index];
    index += sizeof(*(*pph_buffer)->p_scatter_buffer) * SCAN_LINES;

    (*pph_buffer)->p_reflectance_buffer = (uint16_t *)&(*pph_buffer)->p_buffer[index];
    temp_buffer.p_reflectance_buffer    = (uint16_t *)&temp_buffer.p_buffer[index];
    index += sizeof(*(*pph_buffer)->p_reflectance_buffer) * SCAN_LINES;

    (*pph_buffer)->p_rgb_buffer = (uint16_t *)&(*pph_buffer)->p_buffer[index];
    temp_buffer.p_rgb_buffer    = (uint16_t *)&temp_buffer.p_buffer[index];
    index += sizeof(*(*pph_buffer)->p_rgb_buffer) * SCAN_LINES * 3;

    (*pph_buffer)->p_gpu_execution_time = (float *)&(*pph_buffer)->p_buffer[index];
    temp_buffer.p_gpu_execution_time    = (float *)&temp_buffer.p_buffer[index];
    index += sizeof(*(*pph_buffer)->p_gpu_execution_time);

    (*pph_buffer)->p_total_execution_time = (float *)&(*pph_buffer)->p_buffer[index];
    temp_buffer.p_total_execution_time    = (float *)&temp_buffer.p_buffer[index];

    // copy the configuration part of the buffer
    memcpy((*pph_buffer)->p_config, p_config, sizeof(scan_config_t));

    cuda_error = cudaMemcpy(temp_buffer.p_config, p_config, sizeof(scan_config_t), cudaMemcpyHostToDevice);
    if (cuda_error != cudaSuccess)
    {
        LOG_DEBUG("Failed to copy config struct into device buffer\n");
        return RET_ERROR_OTHER;
    }

    // copy the temporary buffer to the device buffer
    cuda_error = cudaMemcpy(*ppd_buffer, &temp_buffer, sizeof(temp_buffer), cudaMemcpyHostToDevice);
    if (cuda_error != cudaSuccess)
    {
        LOG_DEBUG("Failed to copy pointer struct into device buffer\n");
        return RET_ERROR_OTHER;
    }

    return RET_SUCCESS;
}

static void deinit_output_buffers_device(scan_output_buffer_t ** pph_buffer,
                                         scan_output_buffer_t ** ppd_buffer,
                                         uint8_t ** ppd_buffer_raw)
{
    if ((*pph_buffer != NULL))
    {
        if ((*pph_buffer)->p_buffer != NULL)
        {
            cudaFree((*pph_buffer)->p_buffer);
            (*pph_buffer)->p_buffer = NULL;
        }
        cudaFree(*pph_buffer);
        *pph_buffer = NULL;
    }

    if ((*ppd_buffer != NULL))
    {
        cudaFree((*pph_buffer));
        *ppd_buffer = NULL;
    }

    if (*ppd_buffer_raw != NULL)
    {
        cudaFree((*ppd_buffer_raw));
        *ppd_buffer_raw = NULL;
    }
}

#endif


#ifdef STREAM_DATA_OUTPUT_BUFFER_HOST

static ret_code_t init_output_buffers_host(scan_output_buffer_t ** pph_buffer,
                                           scan_output_buffer_t ** ppd_buffer,
                                           uint8_t ** ppd_buffer_raw,
                                           scan_config_t * p_config)
{
    cudaError_t cuda_error;

    cuda_error = cudaHostAlloc(pph_buffer, sizeof(scan_output_buffer_t), cudaHostAllocMapped);
    if (cuda_error != cudaSuccess)
    {
        *pph_buffer = NULL;
        return RET_ERROR_NO_MEM;
    }

    const uint32_t SCAN_LINES_PER_IMAGE = p_config->image_info.num_scan_lines_per_image;
    const uint32_t SCAN_LINES           = p_config->img_processing_config.frames_per_message * SCAN_LINES_PER_IMAGE;

    // first calculate size.
    (*pph_buffer)->buffer_size = 0;

    (*pph_buffer)->buffer_size += sizeof(scan_config_t);
    (*pph_buffer)->buffer_size += sizeof(*(*pph_buffer)->p_height_buffer) * SCAN_LINES;
    (*pph_buffer)->buffer_size += sizeof(*(*pph_buffer)->p_intensity_buffer) * SCAN_LINES;
    (*pph_buffer)->buffer_size += sizeof(*(*pph_buffer)->p_reflectance_buffer) * SCAN_LINES;
    (*pph_buffer)->buffer_size += sizeof(*(*pph_buffer)->p_scatter_buffer) * SCAN_LINES;
    (*pph_buffer)->buffer_size += sizeof(*(*pph_buffer)->p_rgb_buffer) * SCAN_LINES * 3;
    (*pph_buffer)->buffer_size += sizeof(*(*pph_buffer)->p_gpu_execution_time);
    (*pph_buffer)->buffer_size += sizeof(*(*pph_buffer)->p_total_execution_time);

    // then allocate
    cuda_error = cudaHostAlloc(&(*pph_buffer)->p_buffer,(*pph_buffer)->buffer_size, cudaHostAllocMapped);
    if (cuda_error != cudaSuccess)
    {
        (*pph_buffer)->p_buffer = NULL;
        return RET_ERROR_NO_MEM;
    }

    uint32_t index = 0;
    (*pph_buffer)->p_config = (scan_config_t *)&(*pph_buffer)->p_buffer[index];
    index += sizeof(scan_config_t);

    (*pph_buffer)->p_height_buffer = (uint16_t*)&(*pph_buffer)->p_buffer[index];
    index += sizeof(*(*pph_buffer)->p_height_buffer) * SCAN_LINES;

    (*pph_buffer)->p_intensity_buffer = (uint16_t*)&(*pph_buffer)->p_buffer[index];
    index += sizeof(*(*pph_buffer)->p_intensity_buffer) * SCAN_LINES;

    (*pph_buffer)->p_scatter_buffer = (uint16_t*)&(*pph_buffer)->p_buffer[index];
    index += sizeof(*(*pph_buffer)->p_scatter_buffer) * SCAN_LINES;

    (*pph_buffer)->p_reflectance_buffer = (uint16_t*)&(*pph_buffer)->p_buffer[index];
    index += sizeof(*(*pph_buffer)->p_reflectance_buffer) * SCAN_LINES;

    (*pph_buffer)->p_rgb_buffer = (uint16_t*)&(*pph_buffer)->p_buffer[index];
    index += sizeof(*(*pph_buffer)->p_rgb_buffer) * SCAN_LINES * 3;

    (*pph_buffer)->p_gpu_execution_time = (float*)&(*pph_buffer)->p_buffer[index];
    index += sizeof(*(*pph_buffer)->p_gpu_execution_time);

    (*pph_buffer)->p_total_execution_time = (float*)&(*pph_buffer)->p_buffer[index];

    // copy the configuration part of the buffer
    memcpy((*pph_buffer)->p_config, p_config, sizeof(scan_config_t));

    cuda_error = cudaHostGetDevicePointer(ppd_buffer, *pph_buffer,0);
    if (cuda_error != cudaSuccess)
    {
        *ppd_buffer = NULL;
        LOG_DEBUG("Failed to obtain device pointer for output buffer\n");
        return RET_ERROR_OTHER;
    }

    cuda_error = cudaHostGetDevicePointer(ppd_buffer_raw, (*pph_buffer)->p_buffer,0);
    if (cuda_error != cudaSuccess)
    {
        *ppd_buffer_raw = NULL;
        LOG_DEBUG("Failed to obtain device pointer for raw output buffer\n");
        return RET_ERROR_OTHER;
    }

    return RET_SUCCESS;
}

static void deinit_output_buffers_host(scan_output_buffer_t ** pph_buffer,
                                       scan_output_buffer_t ** ppd_buffer,
                                       uint8_t ** ppd_buffer_raw)
{
    if ((*pph_buffer != NULL))
    {
        if ((*pph_buffer)->p_buffer != NULL)
        {
            cudaFree((*pph_buffer)->p_buffer);
            (*pph_buffer)->p_buffer = NULL;
        }
        cudaFree(*pph_buffer);
        *pph_buffer = NULL;
    }

    *ppd_buffer = NULL;
    *ppd_buffer_raw = NULL;
}

#endif


static ret_code_t init_event(cudaEvent_t * p_event)
{
    cudaError_t cuda_error;
    cuda_error = cudaEventCreate(p_event);
    if (cuda_error != cudaSuccess)
    {
        return RET_ERROR_NO_MEM;
    }
    return RET_SUCCESS;
}

static void deinit_event(cudaEvent_t * p_event)
{
    if (*p_event != 0)
    {
        cudaEventDestroy(*p_event);
        *p_event = 0;
    }
}
