#include "tcp_server.h"
#include <stdio.h>
#include <string.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#include <unistd.h>
#include <pthread.h>

#include <signal.h>

#include "log.h"
#include "app_error.h"

static int client_fd = -1;
static int server_fd = -1;
static struct sockaddr_in server;
static uint8_t recv_buffer[RECV_BUFFER_SIZE];
static pthread_t thread;

#define HEADER_SIZE         5


static ret_code_t try_read_length(uint32_t length)
{
    uint32_t bytes_read;
    int temp_bytes_read;

    bytes_read = 0;
    while ( bytes_read < length)
    {
        temp_bytes_read = recv(client_fd,&recv_buffer[bytes_read],length - bytes_read, 0);
        if (temp_bytes_read <= 0)
        {
            return RET_ERROR_OTHER;
        }

        bytes_read += temp_bytes_read;

        pthread_yield();

    }
    return RET_SUCCESS;
}

static void * start_listen(void *args)
{
    struct sockaddr_in  client;
    socklen_t           sock_len = sizeof(struct sockaddr_in);
    tcp_evt_t           evt;
    ret_code_t          ret_code;
    tcp_evt_handler_t   evt_handler = (tcp_evt_handler_t)args;

    while (1)
    {
        client_fd = accept(server_fd,(struct sockaddr *)&client,&sock_len);
        if ( client_fd < 0 )
        {
            continue;
        }

        evt.type = TCP_EVT_TYPE_CONNECTED;
        evt_handler(&evt);

        while (1)
        {
            evt.type = TCP_EVT_TYPE_RECV_MESSAGE;

            // read header
            ret_code = try_read_length(HEADER_SIZE);
            if (ret_code != RET_SUCCESS)
            {
                break;
            }

            // store header information in the event
            evt.message.type = (message_type_t)recv_buffer[0];
            memcpy(&evt.message.request.payload_len, &recv_buffer[1], 4);

            if (evt.message.request.payload_len > RECV_BUFFER_SIZE)
            {
                LOG_DEBUG("Cannot handle this size of request message\n");
                break;
            }

            // read data
            ret_code = try_read_length(evt.message.request.payload_len);
            if (ret_code != RET_SUCCESS)
            {
                break;
            }

            if (evt.message.type == MSG_TYPE_RESPONSE)
            {
                evt.message.response.payload_len    = evt.message.request.payload_len - 2;
                evt.message.response.request_type   = recv_buffer[0];
                evt.message.response.ret_code       = recv_buffer[1];
                evt.message.response.p_payload      = recv_buffer + 2;
            }
            else
            {
                evt.message.request.p_payload = recv_buffer;
            }

            evt_handler(&evt);
        }
        client_fd = -1;
        evt.type = TCP_EVT_TYPE_DISCONNECTED;
        evt_handler(&evt);
    }
}

static ret_code_t create_thread(tcp_server_config_t * p_config)
{
    int err;
    pthread_attr_t attr;

    err = pthread_attr_init(&attr);
    if (err < 0)
    {
        perror("Initializing attrs thread\n");
        close(server_fd);
        return RET_ERROR_OTHER;
    }

    err = pthread_create(&thread, &attr, start_listen, p_config->evt_handler);
    if (err != 0)
    {
        perror("Creating thread\n");
        close(server_fd);
        return RET_ERROR_OTHER;
    }

    return RET_SUCCESS;
}

ret_code_t tcp_server_start_listen(tcp_server_config_t * p_config)
{
    if (p_config == NULL)
    {
        return RET_ERROR_NULL;
    }

    int err;

    memset(&server, 0, sizeof(server));

    server_fd = socket(AF_INET, SOCK_STREAM, 0);
    if (server_fd == -1)
    {
        perror("Creating socket");
        return RET_ERROR_NO_MEM;
    }

    int option = 1;
    err = setsockopt(server_fd, SOL_SOCKET, SO_REUSEADDR, &option, sizeof(option));
    if (err < 0)
    {
        perror("Setting socket options");
        close(server_fd);
        return RET_ERROR_OTHER;
    }

    server.sin_family= AF_INET;
    server.sin_addr.s_addr = INADDR_ANY;
    server.sin_port = htons(p_config->port);

    err = bind(server_fd, (const struct sockaddr*)&server, sizeof(server));
    if (err < 0)
    {
        perror("Binding socket");
        close(server_fd);
        return RET_ERROR_OTHER;
    }

    err = listen(server_fd, 1);
    if (err < 0)
    {
        perror("Listening to socket");
        close(server_fd);
        return RET_ERROR_OTHER;
    }

    // Do no not raise signals when a pipe is broken.
    signal(SIGPIPE, SIG_IGN);

    return create_thread(p_config);
}


ret_code_t tcp_server_send_message(const message_t * p_message)
{
    if (p_message == NULL)
    {
        return RET_ERROR_NULL;
    }

    ret_code_t ret_code;
    uint8_t    header[7]; // only 5 bytes are used when sending a request

    header[0] = p_message->type;

    if (p_message->type != MSG_TYPE_RESPONSE)
    {
        memcpy(&header[1], &p_message->request.payload_len, sizeof(p_message->request.payload_len));

        ret_code = tcp_server_send_data((void*)header, 5);
        RETURN_ON_ERROR(ret_code);

        // send payload if present
        if (p_message->request.payload_len > 0)
        {
            ret_code = tcp_server_send_data((void*)p_message->request.p_payload,p_message->request.payload_len);
        }
    }
    else
    {
        // Temporary increase the payload length to account for the return code
        // and original message type
        uint32_t temp_length = p_message->response.payload_len + 2;

        memcpy(&header[1], &temp_length, sizeof(p_message->response.payload_len));

        header[5] = p_message->response.request_type;
        header[6] = (uint8_t) p_message->response.ret_code;

        ret_code = tcp_server_send_data((void*)header, 7);
        RETURN_ON_ERROR(ret_code);

        // send payload if present
        if (p_message->response.payload_len > 0)
        {
            ret_code = tcp_server_send_data((void*)p_message->response.p_payload,p_message->response.payload_len);
        }
    }

    return ret_code;
}


ret_code_t tcp_server_send_data(void * p_data, uint32_t len)
{
    uint32_t sent_len;

    if (client_fd == -1)
    {
        return RET_ERROR_INVALID_STATE;
    }

    sent_len = send(client_fd, p_data, len, 0);
    if (sent_len == -1)
    {
        perror("Sending data");
        client_fd = -1;
        return RET_ERROR_OTHER;
    }
    else if (sent_len != len)
    {
        return RET_ERROR_OTHER;
    }

    return RET_SUCCESS;
}
