#ifndef TCP_SERVER_H__
#define TCP_SERVER_H__

#ifdef __cplusplus
extern "C"
{
#endif

#include <stdint.h>
#include "ret_code.h"

#define RECV_BUFFER_SIZE 1024  /**< The receive buffer size sets the maximum message size of input input message. */


/** @brief Message type defined by the protocol.
 *
 * @details For more details on the interpretation of each message type, see the
 *          specification of the communication protocol.
 *
 * @note    The protocol uses only one byte for coding the message type. The range is therefore limited to 0-255.
 */
typedef enum
{
    MSG_TYPE_GOTO_STATE                 = 0x01,     /**< Message type for changing the current state. This message is sent from host to device. */
    MSG_TYPE_CONFIG_CAMERA              = 0x02,     /**< Message type for configuring camera parameters. This message is sent from host to device. */
    MSG_TYPE_CONFIG_SCAN                = 0x03,     /**< Message type for configuring the scanner. This message is sent from host to device.  */
    MSG_TYPE_GET_IMAGE                  = 0x04,     /**< Message type for obtaining an image. This message is sent from host to device. */
    MSG_TYPE_SCAN_DATA                  = 0x05,     /**< Message type indicating that it contains line data. This message is sent from device to host. */
    MSG_TYPE_SCAN_ERROR                 = 0x06,     /**< Message type indicating than an error occurred during scanning. The scanner returns to idle. */

    MSG_TYPE_RESPONSE                   = 0x80      /**< Message type indication a response to a request. This message is only sent from device to host. */
} message_type_t;


/** @brief Definition of a message.
 *
 * @details For more details on the structure of a message, see the
 *          specification of the communication protocol.
 */
typedef struct
{
    message_type_t  type;                       /**< The type of the message. Only one byte will be used. */

    union {
        struct
        {
            uint8_t     * p_payload;            /**< Pointer to the data associated with the message type. */
            uint32_t    payload_len;            /**< The length in bytes with the associated data. */
        } request;                              /**< In case of a request, that is @ref message_t::type is @ref MSG_TYPE_GOTO_STATE, @ref MSG_TYPE_CONFIG_CAMERA, @ref MSG_TYPE_CONFIG_IMAGE_PROCESSING, @ref MSG_TYPE_CONFIG_OUTPUT_FORMAT, or @ref MSG_TYPE_GET_IMAGE, the message format is defined by this struct. */

        struct
        {
            message_type_t  request_type;       /**< The request type which the response is related to. Only one byte is used for encoding the request type. */
            ret_code_t      ret_code;           /**< The return code to the request. Only one byte is used for encoding the return code. */
            uint8_t         * p_payload;        /**< Pointer to additional data associated with the response. */
            uint32_t        payload_len;        /**< Length of additional data associated with the response. */
        } response;                             /**< In case of a response, that is @ref message_t::type is @ref MSG_TYPE_RESPONSE, the message format is defined by this struct. */
    };

} message_t;


/** @brief Event type sent from the TCP server.*/
typedef enum
{
    TCP_EVT_TYPE_CONNECTED,                     /**< When a host connects, this event type is sent. */
    TCP_EVT_TYPE_DISCONNECTED,                  /**< When a host disconnect, this event type is sent. */
    TCP_EVT_TYPE_RECV_MESSAGE                   /**< When a message is received, this event type is sent. */
} tcp_evt_type_t;


/** @brief Definition of a TCP server event. */
typedef struct
{
    tcp_evt_type_t  type;                       /**< The type of the event. */
    union
    {
        message_t       message;                    /**< If the type is @ref TCP_EVT_TYPE_RECV_MESSAGE, this is the associated message. */
    };
} tcp_evt_t;


/** @brief Definition of a tcp event handler.
 *
 *  @param[in]  p_evt   Pointer to the received event.
 */
typedef void (*tcp_evt_handler_t)(tcp_evt_t const * p_evt);


/** @brief Configuration parameters used for initializing the TCP server.
 *
 * @details This struct is used for the function @ref tcp_server_start_listen. */
typedef struct
{
    int                 port;           /**< The port were the messages are transmitted through. */
    tcp_evt_handler_t   evt_handler;    /**< Event handler for TCP server events. */
} tcp_server_config_t;


/** @brief Starts the TCP server with the associated configuration.
 *
 *  @param[in]  p_config        The configuration of the TCP server.
 *
 *  @return RET_SUCCESS         If the server was successfully started.
 *  @return RET_ERROR_NULL      If a NULL parameter was provided.
 *  @return RET_ERROR_NO_MEM    If the application failed creating a server socket.
 *  @return RET_ERROR_OTHER     If some internal socket operation failed, or if failed to start the listener thread.
 */
ret_code_t tcp_server_start_listen(tcp_server_config_t * p_config);


/** @brief Sends a TCP message.
 *
 *  @details Depending on the message type, the formatting of the sent message will vary. See the protocol specification for details.
 *
 *  @param[in]  p_message           The message to be sent.
 *
 *  @return RET_SUCCESS             If the message was sent successfully.
 *  @return RET_ERROR_NULL          If a NULL parameter was provided.
 *  @return RET_ERROR_INVALID_PARAM If the type of the provided message is unknown.
 *  @return RET_ERROR_INVALID_STATE If a client is not connected.
 *  @return RET_ERROR_OTHER         If some internal socket operation failed.
 */
ret_code_t tcp_server_send_message(const message_t * p_message);


/** @brief Sends a raw TCP message.
 *
 *  @details Depending on the message type, the formatting of the sent message will vary. See the protocol specification for details.
 *
 *  @param[in]  p_data              Pointer to the raw data to be send.
 *  @param[in]  data_len            The length of the data in bytes to be sent.
 *
 *  @return RET_SUCCESS             If the message was sent successfully.
 *  @return RET_ERROR_INVALID_STATE If a client is not connected.
 *  @return RET_ERROR_OTHER         If some internal socket operation failed.
 */
ret_code_t tcp_server_send_data(void * p_data, uint32_t data_len);


#ifdef __cplusplus
}
#endif

#endif // TCP_SERVER_H__
